# -*- encoding: utf-8 -*-

import time
from astraflux.interface import *


class SnowflakeID:
    """
    A singleton class that generates unique IDs based on the Snowflake algorithm.
    This algorithm ensures that the generated IDs are unique across different machines and time.
    """

    _instance = None
    __initialized = False

    def __new__(cls, *args, **kwargs):
        """
        Overrides the __new__ method to implement the singleton pattern.
        Ensures that only one instance of the Snowflake class is created.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Snowflake: The singleton instance of the Snowflake class.
        """

        if not cls._instance:
            cls._instance = super(SnowflakeID, cls).__new__(cls)
            cls._instance.__initialized = False
        return cls._instance

    def __init__(self, datacenter_id: int = None, machine_id: int = None, sequence: int = 0):
        """
        Initializes the Snowflake ID generator.

        Args:
            datacenter_id (int, optional): The ID of the data center. Defaults to None.
            machine_id (int, optional): The ID of the machine. Defaults to None.
            sequence (int, optional): The initial sequence number. Defaults to 0.
        """
        if self.__initialized:
            return

        self.__initialized = True
        self.ipaddress = get_ipaddr()

        self.start_timestamp = 1288834974657

        self.datacenter_id_bits = 5
        self.machine_id_bits = 5
        self.sequence_bits = 12

        self.max_datacenter_id = (1 << self.datacenter_id_bits) - 1
        self.max_machine_id = (1 << self.machine_id_bits) - 1
        self.max_sequence = (1 << self.sequence_bits) - 1

        self.machine_id_shift = self.sequence_bits
        self.datacenter_id_shift = self.sequence_bits + self.machine_id_bits
        self.timestamp_left_shift = self.sequence_bits + self.machine_id_bits + self.datacenter_id_bits

        if datacenter_id is not None:
            self.datacenter_id = datacenter_id
        else:
            self.datacenter_id = int(self.ipaddress.split('.')[-2])

        if machine_id is not None:
            self.machine_id = machine_id
        else:
            self.machine_id: int = int(self.ipaddress.split('.')[-1])

        self.sequence = sequence
        self.last_timestamp = -1

    @staticmethod
    def _current_timestamp():
        """
        Returns the current timestamp in milliseconds.

        Returns:
            int: The current timestamp in milliseconds.
        """
        return int(time.time() * 1000)

    def _till_next_millis(self, last_timestamp):
        """
        Waits until the next millisecond to ensure the timestamp is greater than the last one.

        Args:
            last_timestamp (int): The last timestamp used to generate an ID.

        Returns:
            int: The new timestamp.
        """
        timestamp = self._current_timestamp()
        while timestamp <= last_timestamp:
            timestamp = self._current_timestamp()
        return timestamp

    @property
    def generate_id(self) -> str:
        """
        Generates a Snowflake ID using the Snowflake algorithm.

        Returns:
            str: A Snowflake ID.
        """
        timestamp = self._current_timestamp()

        if timestamp < self.last_timestamp:
            raise Exception("Clock moved backwards. Refusing to generate id")

        if self.last_timestamp == timestamp:
            self.sequence = (self.sequence + 1) & self.max_sequence
            if self.sequence == 0:
                timestamp = self._till_next_millis(self.last_timestamp)
        else:
            self.sequence = 0

        self.last_timestamp = timestamp

        _id = ((timestamp - self.start_timestamp) << self.timestamp_left_shift) | \
              (self.datacenter_id << self.datacenter_id_shift) | \
              (self.machine_id << self.machine_id_shift) | self.sequence

        return str(_id)


def snowflake_id() -> str:
    """
    Returns a Snowflake ID generator function.
    Returns:
        function: A function that generates Snowflake IDs.
    """
    return SnowflakeID().generate_id


def register():
    from astraflux.interface import snowflake
    snowflake.snowflake_id = snowflake_id

    import sys
    sys.modules['astraflux.interface.snowflake'] = snowflake



# -*- encoding: utf-8 -*-
from astraflux.settings import *
from astraflux.inject import inject_implementation

__all__ = [
    "task_submit_databases",
    "task_submit_databases_and_send",
    "subtask_create",
    "query_task_by_task_id",
    "query_worker_running_number",
    "task_stop",
    "redis_get_task_status_by_task_id"
]


@inject_implementation()
def task_submit_databases(queue: str, message: dict, weight: int = TASK.DEFAULT_VALUE_TASK_WEIGHT) -> str:
    """
        Submit a task to the specified queue.
        Args:
            queue (str): The name of the queue to submit the task to.
            message (dict): The message to be submitted as a task.
            weight (int): The weight of the task. Default is 1.
        Returns:
            str: The task ID associated with the submitted task.
        This method submits the provided task to the specified queue using the RabbitMQ instance.
    """


@inject_implementation()
def task_submit_databases_and_send(queue: str, message: dict, weight: int = TASK.DEFAULT_VALUE_TASK_WEIGHT) -> str:
    """
        Send a message to the queue.
        Args:
            queue (str): The name of the queue to send the message to.
            message (dict): The message to be sent.
            weight (int): The weight of the message. Default is 1.
        Returns:
            str: The task ID associated with the message.
        This method sends the provided message to the specified queue using the RabbitMQ instance.
    """


@inject_implementation()
def subtask_create(source_task_id: str, subtask_queue: str, subtasks: list) -> list:
    """
        Create a subtask for the given task ID.
        Args:
            source_task_id (str): The ID of the task to create a subtask for.
            subtask_queue (str): The name of the queue to create the subtask in.
            subtasks (dict): The subtask to be created.
        Returns:
            str: The subtask ID associated with the created subtask.
        This method creates a subtask for the given task ID using the RabbitMQ instance.
    """


@inject_implementation()
def query_task_by_task_id(task_id: str) -> dict:
    """
    Query the task data of the given task ID.
    Args:
        task_id (str): The ID of the task to query.
    Returns:
        dict: task data
    """


@inject_implementation()
def query_worker_running_number(query: dict):
    """
    Query the number of running workers based on the given query.
    Args:
        query (dict): A dictionary containing the query criteria.
    Returns:
        tuple: A tuple containing the number of running workers and the maximum number of workers.
    """


@inject_implementation()
def task_stop(task_id: str) -> None:
    """
    Stop the given task ID.
    """


@inject_implementation()
def redis_get_task_status_by_task_id(task_id: str) -> str:
    """
    Get the status of the given task ID.
    Args:
        task_id (str): The ID of the task to get the status of.

    Returns:
        str: The status of the given task ID.
    """



# -*- encoding: utf-8 -*-

import abc
import logging

__all__ = ['ServiceConstructor', 'WorkerConstructor']


class Meta(abc.ABC):

    @property
    @abc.abstractmethod
    def task_create(self): ...

    @property
    @abc.abstractmethod
    def task_create_and_run(self): ...

    @property
    @abc.abstractmethod
    def subtask_create(self): ...

    @property
    @abc.abstractmethod
    def loguru(self): ...

    @property
    @abc.abstractmethod
    def name(self): ...


class ServiceMeta(Meta):

    @property
    @abc.abstractmethod
    def service_name(self): ...

    @property
    @abc.abstractmethod
    def service_ipaddr(self): ...

    @property
    @abc.abstractmethod
    def service_version(self): ...


class WorkerMeta(Meta):

    @property
    @abc.abstractmethod
    def worker_name(self): ...

    @property
    @abc.abstractmethod
    def worker_version(self): ...

    @property
    @abc.abstractmethod
    def worker_ipaddr(self): ...

    @abc.abstractmethod
    def run(self, data): ...


class ServiceConstructor(ServiceMeta):
    """
    ServiceConstructor is a class that represents a constructor for a service
    """
    functions: list = []
    service_name: str = None
    service_ipaddr: str = None
    service_version: str = None

    name: str = None
    loguru: logging.Logger = None

    def task_create(self): ...

    def task_create_and_run(self): ...

    def subtask_create(self): ...

    @classmethod
    def setattr(cls, name, value):
        setattr(cls, name, value)

    def __call__(self):
        return self


class WorkerConstructor(WorkerMeta):
    """
    WorkerConstructor is a class that represents a constructor for a worker.
    It also contains a method called run, which takes in a body and runs the worker with the given body.
    """
    functions: list = []

    worker_name: str = None
    worker_ipaddr: str = None
    worker_version: str = None

    name: str = None
    loguru: logging.Logger = None

    def task_create(self): ...

    def task_create_and_run(self): ...

    def subtask_create(self): ...

    @classmethod
    def setattr(cls, name, value):
        setattr(cls, name, value)

    def run(self, body):
        """
        Runs the worker with the given body.
        Args:
            body (dict): The body of the worker.
        """

    def __call__(self):
        return self



# -*- encoding: utf-8 -*-

from astraflux.settings import *
from astraflux.interface.snowflake import snowflake_id
from astraflux.interface.format_time import get_converted_time
from astraflux.interface.rabbitmq import rabbitmq_send_message

from ._mongodb import mongodb_task, mongodb_services
from ._redisdb import redis_task


def _task_required_field_check(message: dict):
    """
    Check if the required fields are present in the task message.
    Args:
        message (dict): The task message.
    Raises:
        Exception: If any of the required fields are missing.
    Returns:
        message (dict): The task message.
    """
    if TASK.KEY_TASK_ID in message and message[TASK.KEY_TASK_ID] is not None and message[TASK.KEY_TASK_ID] != "":
        return message

    message[TASK.KEY_TASK_ID] = snowflake_id()
    return message


def _is_server_run(queue: str) -> bool:
    """
    Check if the given queue is running.
    """
    count = mongodb_services().query_count(query={BUILD.KEY_SERVICE_NAME: queue})
    return count > 0


def _gen_task_message(queue: str, message: dict, weight: int = TASK.DEFAULT_VALUE_TASK_WEIGHT) -> tuple[dict, dict]:
    """
    Generate a task message.
    """

    message = _task_required_field_check(message=message)
    body = {
        TASK.KEY_TASK_BODY: message,
        TASK.KEY_TASK_WEIGHT: weight,
        TASK.KEY_TASK_QUEUE_NAME: queue,
        TASK.KEY_TASK_IS_SUB_TASK: False,
        TASK.KEY_TASK_ID: message[TASK.KEY_TASK_ID],
        TASK.KEY_TASK_STATUS: TASK.KEY_TASK_WAIT_STATUS,
        TASK.KEY_TASK_IS_SUB_TASK_ALL_FINISH: False,
        TASK.KEY_TASK_CREATE_TIME: get_converted_time()
    }
    return body, message


def task_submit_databases(queue: str, message: dict, weight: int = TASK.DEFAULT_VALUE_TASK_WEIGHT) -> str:
    """
        Submit a task to the specified queue.
        Args:
            queue (str): The name of the queue to submit the task to.
            message (dict): The message to be submitted as a task.
            weight (int): The weight of the task. Default is 1.
        Returns:
            str: The task ID associated with the submitted task.
        This method submits the provided task to the specified queue using the RabbitMQ instance.
    """
    if _is_server_run(queue) is False:
        raise f"Server run is not supported {queue}"

    body, message = _gen_task_message(queue=queue, message=message, weight=weight)
    mongodb_task().update_many(query={TASK.KEY_TASK_ID: body[TASK.KEY_TASK_ID]}, update_data=body, upsert=True)
    return message[TASK.KEY_TASK_ID]


def task_submit_databases_and_send(queue: str, message: dict, weight: int = TASK.DEFAULT_VALUE_TASK_WEIGHT) -> str:
    """
        Send a message to the queue.
        Args:
            queue (str): The name of the queue to send the message to.
            message (dict): The message to be sent.
            weight (int): The weight of the message. Default is 1.
        Returns:
            str: The task ID associated with the message.
        This method sends the provided message to the specified queue using the RabbitMQ instance.
    """
    if _is_server_run(queue) is False:
        raise f"Server run is not supported {queue}"

    body, message = _gen_task_message(queue=queue, message=message, weight=weight)

    rabbitmq_send_message(queue=queue, message=message)
    mongodb_task().update_many(query={TASK.KEY_TASK_ID: body[TASK.KEY_TASK_ID]}, update_data=body, upsert=True)
    return message[TASK.KEY_TASK_ID]


def subtask_create(source_task_id: str, subtask_queue: str, subtasks: list) -> list:
    """
        Create a subtask for the given task ID.
        Args:
            source_task_id (str): The ID of the task to create a subtask for.
            subtask_queue (str): The name of the queue to create the subtask in.
            subtasks (dict): The subtask to be created.
        Returns:
            str: The subtask ID associated with the created subtask.
        This method creates a subtask for the given task ID using the RabbitMQ instance.
    """
    if _is_server_run(subtask_queue) is False:
        raise f"Server run is not supported {subtask_queue}"

    subtask_ids = []
    for subtask in subtasks:
        message = _task_required_field_check(message=subtask)
        message[TASK.KEY_TASK_SOURCE_ID] = source_task_id
        body = {
            TASK.KEY_TASK_BODY: message,
            TASK.KEY_TASK_WEIGHT: TASK.DEFAULT_VALUE_TASK_WEIGHT,
            TASK.KEY_TASK_QUEUE_NAME: subtask_queue,
            TASK.KEY_TASK_SOURCE_ID: source_task_id,
            TASK.KEY_TASK_IS_SUB_TASK: True,
            TASK.KEY_TASK_ID: message[TASK.KEY_TASK_ID],
            TASK.KEY_TASK_STATUS: TASK.KEY_TASK_WAIT_STATUS,
            TASK.KEY_TASK_IS_SUB_TASK_ALL_FINISH: False,
            TASK.KEY_TASK_CREATE_TIME: get_converted_time()
        }
        subtask_ids.append(message[TASK.KEY_TASK_ID])
        mongodb_task().update_many(query={TASK.KEY_TASK_ID: body[TASK.KEY_TASK_ID]}, update_data=body, upsert=True)
    return subtask_ids


def query_task_by_task_id(task_id: str) -> dict:
    """
    Query the task data of the given task ID.
    Args:
        task_id (str): The ID of the task to query.
    Returns:
        dict: task data
    """
    data = mongodb_task().query_all(
        query={TASK.KEY_TASK_ID: task_id},
        field={'_id': 0}
    )
    data = [i for i in data]
    return data[0] if len(data) > 0 else {}


def query_worker_running_number(query: dict):
    """
    Query the number of running workers based on the given query.
    Args:
        query (dict): A dictionary containing the query criteria.
    Returns:
        tuple: A tuple containing the number of running workers and the maximum number of workers.
    """

    data = mongodb_services().query_all(
        query=query,
        field={
            '_id': 0,
            BUILD.KEY_WORKER_RUN_PROCESS: 1,
            BUILD.KEY_WORKER_MAX_PROCESS: 1
        }
    )
    if len(data) > 0:
        return len(data[0].get(BUILD.KEY_WORKER_RUN_PROCESS)), data[0].get(BUILD.KEY_WORKER_MAX_PROCESS)
    return 0, 0


def task_stop(task_id: str) -> None:
    """
    Stop the given task ID.
    """
    mongodb_task().update_many(
        query={TASK.KEY_TASK_ID: task_id},
        update_data={TASK.KEY_TASK_STATUS: TASK.KEY_TASK_STOP_STATUS}
    )
    mongodb_task().update_many(
        query={TASK.KEY_TASK_SOURCE_ID: task_id},
        update_data={TASK.KEY_TASK_STATUS: TASK.KEY_TASK_STOP_STATUS}
    )

    subtask_ids = mongodb_task().query_all(
        query={TASK.KEY_TASK_SOURCE_ID: task_id},
        field={TASK.KEY_TASK_ID: 1, '_id': 0}
    )

    for subtask_id in subtask_ids:
        redis_task().set_hash_value(
            subtask_id.get(TASK.KEY_TASK_ID),
            {TASK.KEY_TASK_STATUS: TASK.KEY_TASK_STOP_STATUS}, expire=604800)

    redis_task().set_hash_value(task_id, {TASK.KEY_TASK_STATUS: TASK.KEY_TASK_STOP_STATUS}, expire=604800)


def redis_get_task_status_by_task_id(task_id: str) -> str:
    """
    Get the status of the given task ID.
    Args:
        task_id (str): The ID of the task to get the status of.

    Returns:
        str: The status of the given task ID.
    """
    data = redis_task().get_hash_value(key=task_id)
    return data.get(TASK.KEY_TASK_STATUS)


def register():
    from astraflux.interface import databases_api
    databases_api.task_submit_databases = task_submit_databases
    databases_api.task_submit_databases_and_send = task_submit_databases_and_send
    databases_api.subtask_create = subtask_create
    databases_api.query_task_by_task_id = query_task_by_task_id
    databases_api.query_worker_running_number = query_worker_running_number
    databases_api.task_stop = task_stop
    databases_api.redis_get_task_status_by_task_id = redis_get_task_status_by_task_id

    import sys
    sys.modules['astraflux.interface.databases_api'] = databases_api



# -*- encoding: utf-8 -*-
import os
import sys
import argparse

from astraflux.inject import inject_init
from astraflux.servers.build import Build

from astraflux.settings import *
from astraflux.interface import *


class RunServer:

    def __init__(self, config, cls_path):
        self.config = config
        self.cls_path = cls_path

    def server_start(self):
        build = Build(
            config=self.config,
            cls_path=self.cls_path,
            build_type='service',
            constructor=ServiceConstructor
        )
        constructor = build.build(task_id=None)

        service_data = {
            BUILD.KEY_NAME: constructor.name,
            BUILD.KEY_SERVICE_IPADDR: constructor.service_ipaddr,
            BUILD.KEY_SERVICE_NAME: constructor.service_name,
            BUILD.KEY_SERVICE_VERSION: constructor.service_version,
            BUILD.KEY_SERVICE_PID: os.getpid(),
            BUILD.KEY_SERVICE_FUNCTIONS: constructor.functions
        }

        mongodb_services().update_many(
            query={
                BUILD.KEY_SERVICE_IPADDR: constructor.service_ipaddr,
                BUILD.KEY_SERVICE_NAME: constructor.service_name
            },
            update_data=service_data,
            upsert=True
        )

        constructor.loguru.info('Service started == {}'.format(service_data))

        service_running(
            service_cls=constructor,
            config={RABBITMQ.KEY_RABBITMQ_URI: self.config.get(RABBITMQ.KEY_RABBITMQ_URI)}
        )


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="run service script")

    parser.add_argument("--config", type=str, help="service config")
    parser.add_argument("--path", type=str, help="service path")
    args = parser.parse_args()

    inject_init()

    configs = load_config(args.config)

    sys.path.append(configs[KEY_ROOT_PATH])

    RunServer(config=configs, cls_path=args.path).server_start()



# -*- encoding: utf-8 -*-

from astraflux.inject import inject_implementation

__all__ = ['snowflake_id']


@inject_implementation()
def snowflake_id() -> str:
    """
    Returns a Snowflake ID generator function.
    Returns:
        function: A function that generates Snowflake IDs.
    """



# -*- encoding: utf-8 -*-
import redis
from astraflux.settings.keys import *

_REDIS_CONFIG = REDIS.DEFAULT_VALUE_REDIS_URI


class RedisClient:
    """
    Client to interact with a Redis database.
    """
    _instance = None

    def __init__(self, db):
        url = '{}/{}'.format(_REDIS_CONFIG, db)
        _connection_pool = redis.ConnectionPool.from_url(url=url)
        self._client = redis.Redis(connection_pool=_connection_pool)

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(RedisClient, cls).__new__(cls)
        return cls._instance

    def set_hash_value(self, key, value, expire: int):
        """
        Set a hash value for a key.
        """
        self._client.hmset(key, value)
        if expire > 0:
            self._client.expire(key, expire)

    def get_hash_value(self, key):
        """
        Get a hash value for a key.
        """
        data = self._client.hgetall(key)
        return {k.decode(): v.decode() for k, v in data.items()}

    def update_expire(self, key, expire: int):
        """
        Update the expiration time for a given key.
        """
        self._client.expire(key, expire)

    def delete(self, key):
        """
        Delete a key.
        """
        self._client.delete(key)

    def get_db_all_value(self):
        """
        Get all values from a Redis database.
        """
        all_keys = self._client.keys('*')
        result = {}
        for key in all_keys:
            result[key] = self.get_hash_value(key)
        return result


def initialization_redis(config: dict):
    """
    Initialize a Redis client for use with astraflux
    """
    global _REDIS_CONFIG
    _REDIS_CONFIG = config.get(REDIS.KEY_REDIS_URI, REDIS.DEFAULT_VALUE_REDIS_URI)


def redis_task() -> RedisClient:
    """
    Generate a Redis task with a REDIS configuration.
    """
    return RedisClient(0)


def redis_services() -> RedisClient:
    """
    Generate a Redis service with a REDIS configuration
    """
    return RedisClient(1)


def register():
    from astraflux.interface import redisdb
    redisdb.redis_task = redis_task
    redisdb.redis_services = redis_services
    redisdb.initialization_redis = initialization_redis

    import sys
    sys.modules['astraflux.interface.redisdb'] = redisdb



# -*- encoding: utf-8 -*-
from astraflux.inject import inject_implementation

__all__ = [
    "get_date_time_obj",
    "format_converted_time",
    "get_converted_time",
    "get_yes_today",
    "get_yesterday_date",
    "convert_timestamp_to_timezone",
    "get_converted_timestamp",
    "get_date_list",
    "get_week_num",
    "get_current_week",
    "is_timestamp_within_days",
    "convert_timestamp_to_timezone_obj",
    "convert_timestamp_to_timezone_str",
    'get_converted_time_float'
]


@inject_implementation()
def get_date_time_obj(data_str: str, fmt=False, timezone=False):
    """
    Specify the timezone and format, and return a time object.

    Args:
        data_str (str): The time string to be converted.
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        datetime.datetime: A datetime object representing the converted time.
    """


@inject_implementation()
def format_converted_time(data_str: str, fmt=False, timezone=False, r_fmt=False):
    """
    Format a time string according to the specified format and timezone.

    Args:
        data_str (str): The time string to be formatted.
        fmt (str or bool): The input format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.
        r_fmt (str or bool): The output format of the time string. If False, use the input format.

    Returns:
        str: A formatted time string.
    """


@inject_implementation()
def get_converted_time(fmt=False, timezone=False):
    """
    Specify timezone and format, return the current time.

    Args:
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        str: A string representing the current time in the specified format and timezone.
    """


@inject_implementation()
def get_converted_time_float(fmt=False, timezone=False):
    """
    Specify timezone and format, return the current time as a float.
    Args:
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.
    Returns:
        float: A float representing the current time in the specified format and timezone.
    """


@inject_implementation()
def get_yes_today(data_str: str, fmt=False, timezone=False):
    """
    Get the previous day's date from the given date string.

    Args:
        data_str (str): The date string.
        fmt (str or bool): The format of the date string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        str: A string representing the previous day's date in the specified format and timezone.
    """


@inject_implementation()
def get_yesterday_date(fmt=False, timezone=False, days=1):
    """
    Specify timezone and format, return the previous day's time.

    Args:
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.
        days (int): The number of days to subtract from the current date. Default is 1.

    Returns:
        str: A string representing the previous day's time in the specified format and timezone.
    """


@inject_implementation()
def convert_timestamp_to_timezone(timestamp, fmt=False, timezone=False):
    """
    Convert a timestamp to a time string in the specified timezone and format.

    Args:
        timestamp (float): The timestamp to be converted.
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        str: A string representing the converted time in the specified format and timezone.
    """


@inject_implementation()
def get_converted_timestamp(date_string: str, fmt=False, timezone=False):
    """
    Convert a time string to a timestamp in the specified timezone and format.

    Args:
        date_string (str): The time string to be converted.
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        float: A timestamp representing the converted time.
    """


@inject_implementation()
def get_date_list(start_day: str, end_day: str):
    """
    Obtain every day within the time range.

    Args:
        start_day (str): The start date in the format '%Y%m%d'.
        end_day (str): The end date in the format '%Y%m%d'.

    Returns:
        list: A list of strings representing each day within the time range in the format '%Y%m%d'.
    """


@inject_implementation()
def get_week_num(date_str: str):
    """
    Obtain a week number from the given date string.
    Args:
        date_str (str): The date string in the format '%Y-%m-%d'.
    Returns:
        int: The week number.
    """


@inject_implementation()
def get_current_week():
    """
    Obtain current week number.
    Returns:
        int: The week number.
    """


@inject_implementation()
def is_timestamp_within_days(timestamp: int, fmt=False, timezone=False):
    """
    Check if a timestamp is within days.
    Args:
        timestamp (int): The timestamp to be checked.
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.
    Returns:
        bool: True if the timestamp is within days, False otherwise.
    """


@inject_implementation()
def convert_timestamp_to_timezone_obj(timestamp, timezone=False):
    """
    Convert a timestamp to a time string.
    Args:
        timestamp (int): The timestamp to be converted.
        timezone (str or bool): The timezone. If False, use the default timezone.
    Returns:
        datetime.datetime: A datetime object representing the converted time.
    """


@inject_implementation()
def convert_timestamp_to_timezone_str(timestamp, timezone=False, fmt=False):
    """
    Convert a timestamp to a time string.
    Args:
        timestamp (int): The timestamp to be converted.
        timezone (str or bool): The timezone. If False, use the default timezone.
        fmt (str or bool): The format of the time string. If False, use the default format.
    Returns:
        str: A string representing the converted time in the specified format and timezone.
    """



# -*- encoding: utf-8 -*-

from astraflux.inject import inject_implementation

__all__ = [
    'initialization_redis',
    'redis_task',
    'redis_services'
]


class RedisClient:
    """
    Client to interact with a Redis database.
    """

    def set_hash_value(self, key, value, expire: int):
        """
        Set a hash value for a key.
        """

    def get_hash_value(self, key):
        """
        Get a hash value for a key.
        """

    def update_expire(self, key, expire: int):
        """
        Update the expiration time for a given key.
        """

    def delete(self, key):
        """
        Delete a key.
        """

    def get_db_all_value(self):
        """
        Get all values from a Redis database.
        """


@inject_implementation()
def initialization_redis(config: dict):
    """
    Initialize a Redis client for use with astraflux
    """


@inject_implementation()
def redis_task() -> RedisClient:
    """
    Generate a Redis task with a REDIS configuration.
    """


@inject_implementation()
def redis_services() -> RedisClient:
    """
    Generate a Redis service with a REDIS configuration
    """



# -*- encoding: utf-8 -*-
from astraflux.inject import inject_implementation

__all__ = [
    'initialization_rpc_proxy',
    'generate_unique',
    'remote_call',
    'proxy_call',
    'rpc_decorator',
    'service_running'
]


@inject_implementation()
def initialization_rpc_proxy(config: dict):
    """
    Initialize the logger with the given configuration.
    Args:
        config (dict): A dictionary containing the configuration.
    """


@inject_implementation()
def generate_unique():
    """
    Generates a unique identifier.
    Returns:
        str: The generated identifier.
    """


@inject_implementation()
def remote_call(service_name: str, method_name: str, **params):
    """
    Makes a remote procedure call to the specified service and method with the given parameters.

    Args:
        service_name (str): The name of the service to call.
        method_name (str): The name of the method to call.
        **params: Arbitrary keyword arguments to pass to the method.

    Returns:
        Any: The result of the remote procedure call.
    """


@inject_implementation()
def proxy_call(service_name: str, method_name: str, **params):
    """
    Makes a remote procedure call to the specified service and method with the given parameters.

    Args:
        service_name (str): The name of the service to call.
        method_name (str): The name of the method to call.
        **params: Arbitrary keyword arguments to pass to the method.

    Returns:
        Any: The result of the remote procedure call.
    """


@inject_implementation()
def rpc_decorator(func):
    """
    Decorator for RPC functions.
    Args:
        func (function): The function to be decorated.
    Returns:
        function: The decorated function.
    """


@inject_implementation()
def service_running(service_cls, config):
    """
    Start a RabbitMQ consumer.
    Args:
        config (dict): The AMQP URL for the RabbitMQ server.
        service_cls (class): The function to be called when a message is received.
    """



# -*- encoding: utf-8 -*-

from astraflux.inject import inject_implementation

__all__ = ['load_settings', 'current_dir', 'global_config', 'load_config', 'dump_config']


@inject_implementation()
def current_dir() -> str:
    """
    Return the current directory of the current working directory.
    """


@inject_implementation()
def global_config() -> dict:
    """
    Return the global configuration settings.
    """


@inject_implementation()
def load_settings(yaml_file: str, workspace: str) -> None:
    """
    Load settings from yaml file.
    :param yaml_file: yaml file path
    :param workspace: workspace path
    :return:
    """


@inject_implementation()
def load_config(encoded_config: str) -> dict:
    """
    Decode and load the encoded configuration string.

    Args:
        encoded_config (str): The encoded configuration string.

    Returns:
        dict: The decoded configuration dictionary.
    """


@inject_implementation()
def dump_config(config: dict) -> str:
    """
    Dump the encoded configuration string.
    """



# -*- encoding: utf-8 -*-

from .rpc import *
from .run import *
from .logger import *
from .redisdb import *
from .network import *
from .mongodb import *
from .rabbitmq import *
from .snowflake import *
from .scheduler import *
from .async_task import *
from .format_time import *
from .databases_api import *



# -*- encoding: utf-8 -*-

import time
import pika
import pickle
import builtins
from pika.exceptions import ChannelClosed

from astraflux.settings import *
from astraflux.interface import *


class ServiceUnavailableError(Exception):

    def __init__(self, service_name):
        super().__init__(f"Service '{service_name}' is not available")
        self.service_name = service_name


class RpcClient:
    """
    A RabbitMQ RPC client for making remote procedure calls.
    Args:
        config (dict): A dictionary containing the RabbitMQ configuration.
            - 'RABBITMQ_URI' (str): The RabbitMQ URI in the format 'amqp://user:password@host:port'.
    Attributes:
        config (dict): The RabbitMQ configuration.
        host (str): The RabbitMQ host.
        port (int): The RabbitMQ port.
        user (str): The RabbitMQ username.
        passwd (str): The RabbitMQ password.
        credentials (pika.PlainCredentials): The RabbitMQ credentials.
        connection (pika.BlockingConnection): The RabbitMQ connection.
        channel (pika.BlockingChannel): The RabbitMQ channel.
        queue (str): The RabbitMQ queue name
    """

    def __init__(self, config: dict):
        self.config = config
        self.timeout = config.get(RPC.KEY_RPC_CALL_TIMEOUT, RPC.DEFAULT_VALUE_RPC_CALL_TIMEOUT)

        self._validate_config()

        self.response = None
        self.corr_id = snowflake_id()

        # get host, port, user, passwd
        self.host, self.port, self.user, self.passwd = self._parse_config()
        self.credentials = pika.PlainCredentials(self.user, self.passwd)

        # create connection and channel
        self.connection = self._create_connection()
        self.channel = self.connection.channel()

        # create queue
        self.queue = self.channel.queue_declare(
            queue='',
            exclusive=True,
            auto_delete=True
        ).method.queue

        # consume response
        self.channel.basic_consume(
            queue=self.queue,
            on_message_callback=self._on_response,
            auto_ack=True
        )

    def _validate_config(self):
        if not self.config.get(RABBITMQ.KEY_RABBITMQ_URI):
            raise ValueError("Missing RabbitMQ URI in config")

    def _create_connection(self):
        """
        Create a RabbitMQ connection using the provided configuration.
        Returns:
            pika.BlockingConnection: The RabbitMQ connection.
        """
        params = pika.ConnectionParameters(
            host=self.host,
            port=self.port,
            credentials=self.credentials,
            heartbeat=600,
            connection_attempts=3,
            retry_delay=5
        )
        return pika.BlockingConnection(params)

    def _check_service_available(self, service_name):
        try:
            self.channel.queue_declare(
                queue=service_name,
                passive=True
            )
            return True
        except ChannelClosed as e:
            if e.args[0] == 404:
                return False
            raise

    @staticmethod
    def _raise_rpc_exception(error_info):
        """
        Raises an exception based on the provided error information.
        Args:
            error_info (dict): A dictionary containing error information.
                - 'type' (str): The type of the exception (e.g., 'TypeError', 'ValueError').
                - 'exception' (str): The exception message.
        Raises:
            TypeError: If the 'type' is 'TypeError'.
            ValueError: If the 'type' is 'ValueError'.
            KeyError: If the 'type' is 'KeyError'.
            AttributeError: If the 'type' is 'AttributeError'.
            RuntimeError: If the 'type' is not recognized or if there is an issue with the exception.
        """
        ex_type = error_info.get('type', 'RpcError')
        ex_msg = error_info.get('exception', 'Unknown RPC error')

        allowed_exceptions = [
            'TypeError', 'ValueError', 'KeyError',
            'AttributeError', 'RuntimeError', 'PermissionError'
        ]

        try:
            if ex_type in allowed_exceptions:
                exception_class = getattr(builtins, ex_type)
                if issubclass(exception_class, Exception):
                    raise exception_class(ex_msg)
            raise RuntimeError(ex_msg)

        except AttributeError:
            raise RuntimeError(ex_msg)

    def _parse_config(self):
        """
        Parse the RabbitMQ configuration string to extract host, port, user and password.

        Returns:
            tuple: A tuple containing host (str), port (int), user (str) and password (str).
        """

        parts = self.config.get(RABBITMQ.KEY_RABBITMQ_URI).split('@')

        user_passwd = parts[0].split('//')[1]
        host_port = parts[1]

        user, passwd = user_passwd.split(':')
        host, port = host_port.split(':')

        return host, int(port), user, passwd

    def _on_response(self, ch, method, props, body):
        if self.corr_id == props.correlation_id:
            self.response = pickle.loads(body)

    def call(self, service_name, method_name, *args, **kwargs):
        """
        Call a remote procedure on the specified service.
        Args:
            service_name (str): The name of the service to call.
            method_name (str): The name of the method to call on the service.
            *args: Positional arguments to pass to the method.
            **kwargs: Keyword arguments to pass to the method.
        Returns:
            Any: The result of the method call.
        Raises:
            ServiceUnavailableError: If the specified service is not available.
            TimeoutError: If the RPC call times out.
        """

        if not self._check_service_available(service_name):
            raise ServiceUnavailableError(service_name)

        request = {
            'method': method_name,
            'args': args,
            'kwargs': kwargs
        }

        self.channel.basic_publish(
            exchange='',
            routing_key=service_name,
            properties=pika.BasicProperties(
                reply_to=self.queue,
                correlation_id=self.corr_id,
                delivery_mode=2
            ),
            body=pickle.dumps(request)
        )

        start_time = time.time()
        while self.response is None:
            if time.time() - start_time > self.timeout:
                raise TimeoutError("RPC call timed out")
            self.connection.process_data_events()

        if isinstance(self.response, dict):
            status = self.response.get('status')
            if status == 'error':
                self._raise_rpc_exception(self.response)
            if status == 'success':
                return self.response.get('result')
        return self.response



# -*- encoding: utf-8 -*-

import time
import threading
import multiprocessing
from typing import Callable, Optional, Dict, Any, List
import uuid


class BaseTask:
    """Base class for all tasks"""

    def __init__(self, task_id: str, target: Callable, args: tuple = (), kwargs: dict = None):
        self.task_id = task_id
        self.target = target
        self.args = args or ()
        self.kwargs = kwargs or {}
        self._is_running = False
        self._stop_event = None
        self.result = None
        self.exception = None

    def start(self):
        """start the task"""
        raise NotImplementedError

    def stop(self):
        """stop the task"""
        if self._stop_event:
            self._stop_event.set()

    def is_running(self) -> bool:
        """check if task is running"""
        return self._is_running

    def get_result(self, task_id=None):
        """get result from task"""
        return self.result

    def get_exception(self):
        """get exception from task"""
        return self.exception

    def wait(self, timeout: Optional[float] = None) -> bool:
        """wait for task to finish"""
        raise NotImplementedError


class ThreadTask(BaseTask):
    """asynchronous task"""

    def __init__(self, task_id: str, target: Callable, args: tuple = (), kwargs: dict = None):
        super().__init__(task_id, target, args, kwargs)
        self._thread = None
        self._stop_event = threading.Event()

    def _run(self):
        """task is running"""
        try:
            self._is_running = True
            self.result = self.target(*self.args, **self.kwargs)
        except Exception as e:
            self.exception = e
        finally:
            self._is_running = False

    def start(self):
        if self._is_running:
            return False

        self._thread = threading.Thread(target=self._run)
        self._thread.daemon = True
        self._thread.start()
        return True

    def stop(self):
        super().stop()
        if self._thread and self._thread.is_alive():
            self._thread.join(timeout=1.0)

        self._is_running = False

    def wait(self, timeout: Optional[float] = None) -> bool:
        if not self._thread:
            return True
        self._thread.join(timeout=timeout)
        return not self._thread.is_alive()


class ProcessTask(BaseTask):
    """asynchronous task"""

    def __init__(self, task_id: str, target: Callable, args: tuple = (), kwargs: dict = None):
        super().__init__(task_id, target, args, kwargs)
        self._process = None
        self._stop_event = multiprocessing.Event()
        self._queue = multiprocessing.Queue()

    def _run(self, _queue):
        """task is running"""
        try:
            self._is_running = True
            result = self.target(*self.args, **self.kwargs)
            _queue.put(result)
        except Exception as e:
            _queue.put(e)
        finally:
            self._is_running = False

    def start(self):
        if self._is_running:
            return False

        self._process = multiprocessing.Process(target=self._run, args=(self._queue,))
        self._process.daemon = True
        self._process.start()

        item = self._queue.get()
        self.result = item

        return True

    def stop(self):
        super().stop()
        if self._process and self._process.is_alive():
            self._process.terminate()
        self._is_running = False

    def wait(self, timeout: Optional[float] = None) -> bool:
        if not self._process:
            return True

        self._process.join(timeout=timeout)

        return not self._process.is_alive()


class TaskManager:
    """asynchronous task manager"""

    _instance = None
    tasks: Dict[str, BaseTask] = {}
    task_lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(TaskManager, cls).__new__(cls)
        return cls._instance

    def add_task(
            self,
            task_type: str,
            target: Callable,
            task_id: Optional[str] = None,
            args: tuple = (),
            kwargs: dict = None) -> str:
        """
        add task
        :param task_type: ('thread' 或 'process')
        :param target: task function
        :param task_id: optional task ID
        :param args: task arguments
        :param kwargs: task keyword arguments
        :return: task_id
        """
        if task_id is None:
            task_id = str(uuid.uuid4())

        with self.task_lock:
            if task_id in self.tasks:
                raise ValueError(f"Task '{task_id}' already exists. Skipping insertion.")

            if task_type == 'thread':
                task = ThreadTask(task_id, target, args, kwargs)
            elif task_type == 'process':
                task = ProcessTask(task_id, target, args, kwargs)
            else:
                raise ValueError(f"Invalid task type: '{task_type}'")

            self.tasks[task_id] = task
            return task_id

    def run_task(self, task_id: str) -> bool:
        """run task"""
        with self.task_lock:
            task = self.tasks.get(task_id)
            if not task:
                return False
            return task.start()

    def stop_task(self, task_id: str) -> bool:
        """stop task"""
        with self.task_lock:
            task = self.tasks.get(task_id)
            if not task:
                return False
            task.stop()
            return True

    def remove_task(self, task_id: str) -> bool:
        """remove task"""
        with self.task_lock:
            task = self.tasks.get(task_id)
            if not task:
                return False

            if task.is_running():
                task.stop()

            del self.tasks[task_id]
            return True

    def get_task_status(self, task_id: str) -> Optional[Dict[str, Any]]:
        """get task status"""
        with self.task_lock:
            task = self.tasks.get(task_id)
            if not task:
                return None

            return {
                'id': task.task_id,
                'running': task.is_running(),
                'result': task.get_result(task_id),
                'exception': task.get_exception()
            }

    def wait_task(self, task_id: str, timeout: Optional[float] = None) -> bool:
        """wait task"""
        task = self.tasks.get(task_id)
        if not task:
            return False
        return task.wait(timeout)

    def stop_all_tasks(self):
        """stop all tasks"""
        with self.task_lock:
            for task in self.tasks.values():
                if task.is_running():
                    task.stop()

    def list_tasks(self) -> Dict[str, Dict[str, Any]]:
        """get all tasks"""
        with self.task_lock:
            return {
                task_id: self.get_task_status(task_id)
                for task_id in self.tasks
            }

    def run_all_tasks(self) -> List[str]:
        """
        run all tasks
        :return: list of started task IDs
        """
        started_tasks = []
        with self.task_lock:
            for task_id, task in self.tasks.items():
                if not task.is_running() and task.start():
                    started_tasks.append(task_id)
        return started_tasks

    def wait_all_tasks(self, timeout: Optional[float] = None) -> bool:
        """
        wait all tasks
        :param timeout: maximum time to wait (seconds)
        :return: True if all tasks completed, False if timeout
        """
        start_time = time.time()
        all_completed = False

        while not all_completed:
            all_completed = True
            with self.task_lock:
                for task in self.tasks.values():
                    if task.is_running():
                        all_completed = False
                        break

            if not all_completed:
                if timeout is not None and (time.time() - start_time) > timeout:
                    return False
                time.sleep(0.1)

        return True


def async_task_add(
        task_type: str,
        target: Callable,
        args: tuple = (),
        kwargs: dict = None,
        task_id: Optional[str] = None) -> str:
    """
    add task
    :param task_type: ('thread' 或 'process')
    :param target: task function
    :param args: task arguments
    :param kwargs: task keyword arguments
    :param task_id: optional task ID
    :return: task_id
    """
    return TaskManager().add_task(
        task_type=task_type,
        target=target,
        args=args,
        kwargs=kwargs,
        task_id=task_id
    )


def async_task_run(task_id: str) -> bool:
    """run task"""
    return TaskManager().run_task(task_id=task_id)


def async_task_get_status(task_id: str) -> Optional[Dict[str, Any]]:
    """get task status"""
    return TaskManager().get_task_status(task_id=task_id)


def async_task_wait(task_id: str, timeout: Optional[float] = None) -> bool:
    """wait task"""
    return TaskManager().wait_task(task_id=task_id, timeout=timeout)


def async_task_stop(task_id: str) -> bool:
    """stop a specific task"""
    return TaskManager().stop_task(task_id=task_id)


def async_task_remove(task_id: str) -> bool:
    """remove a specific task"""
    return TaskManager().remove_task(task_id=task_id)


def async_task_stop_all():
    """stop all tasks"""
    TaskManager().stop_all_tasks()


def async_task_list() -> Dict[str, Dict[str, Any]]:
    """get all tasks"""
    return TaskManager().list_tasks()


def async_task_run_all() -> List[str]:
    """
    run all tasks
    :return: list of started task IDs
    """
    return TaskManager().run_all_tasks()


def async_task_wait_all(timeout: Optional[float] = None) -> bool:
    """
    wait all tasks
    :param timeout: maximum time to wait (seconds)
    :return: True if all tasks completed, False if timeout
    """
    return TaskManager().wait_all_tasks(timeout=timeout)


def register():
    from astraflux.interface import async_task
    async_task.async_task_add = async_task_add
    async_task.async_task_run = async_task_run
    async_task.async_task_get_status = async_task_get_status
    async_task.async_task_wait = async_task_wait
    async_task.async_task_stop_all = async_task_stop_all
    async_task.async_task_list = async_task_list
    async_task.async_task_run_all = async_task_run_all
    async_task.async_task_wait_all = async_task_wait_all

    import sys
    sys.modules['astraflux.interface.async_task'] = async_task



# -*- encoding: utf-8 -*-

import time
import pytz
import pickle
import threading
import multiprocessing
from typing import Dict
from pymongo import MongoClient
from threading import Thread, Event
from datetime import datetime, timedelta

from astraflux.settings import *
from astraflux.interface import *
from .cronparser import CronParser


class Scheduler:
    _instance = None

    _db = None
    _jobs = None
    _locks = None
    _client = None
    _thread = None
    _running = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Scheduler, cls).__new__(cls)
            cls._instance.__initialize(*args, **kwargs)
        return cls._instance

    def __initialize(self, config: dict):
        self.logger = loguru(filename=KEY_PROJECT_NAME, task_id='scheduler')
        self.local_ip = get_ipaddr()
        self._client = MongoClient(config[MONGODB.KEY_MONGODB_URI])
        self._db = self._client[KEY_PROJECT_NAME]
        self._jobs = self._db.scheduled_jobs
        self._locks = self._db.job_locks
        self._running = Event()
        self._thread = None

        self.lock_refresh_interval = 5
        self.lock_expire_seconds = 10
        self.lock_refresh_threads: Dict[str, Event] = {}
        self.lock_thread_lock = threading.Lock()

    def start(self):
        self._running.set()
        self._thread = Thread(target=self._run_loop)
        self._thread.start()

    def stop(self):
        self._running.clear()
        if self._thread:
            self._thread.join()

        with self.lock_thread_lock:
            for job_id, stop_event in self.lock_refresh_threads.items():
                stop_event.set()
            self.lock_refresh_threads.clear()

    def _run_loop(self):
        while self._running.is_set():
            try:
                now = datetime.now(pytz.utc)
                due_jobs = self._jobs.find({
                    "enabled": True,
                    "next_run_time": {"$lte": now}
                })
                for job in due_jobs:
                    self._execute_job(job)
                time.sleep(1)
            except Exception as e:
                self.logger.error(f"Error in scheduler loop: {str(e)}")
                time.sleep(5)

    def _execute_job(self, job):
        ipaddrs = job.get("ipaddrs", [])
        if ipaddrs and self.local_ip not in ipaddrs:
            self.logger.debug(f"Skip task {job['_id']}, the current IP {self.local_ip} is not in the execution list")
            return

        lock_expire = datetime.now(pytz.utc) + timedelta(seconds=self.lock_expire_seconds)
        lock_result = self._locks.update_one(
            {"job_id": job["_id"], "expire_at": {"$lt": datetime.now(pytz.utc)}},
            {"$set": {"expire_at": lock_expire}},
            upsert=True
        )

        if lock_result.modified_count > 0 or lock_result.upserted_id:
            try:
                func = pickle.loads(job["func"])
                exec_type = job.get("exec_type", "thread")

                self._start_lock_refresh(job["_id"])

                if exec_type == "process":
                    self.logger.debug(f"Starting process task: {job['_id']}")
                    process = multiprocessing.Process(
                        target=self._run_task,
                        args=(func, job["args"], job["kwargs"])
                    )
                    process.daemon = True
                    process.start()
                else:
                    self.logger.debug(f"Starting thread task: {job['_id']}")
                    thread = Thread(
                        target=self._run_task,
                        args=(func, job["args"], job["kwargs"])
                    )
                    thread.daemon = True
                    thread.start()

                cron_parser = CronParser(
                    cron_str=job["cron"],
                    timezone=job.get("timezone", "UTC")
                )
                next_run = cron_parser.get_next_run(datetime.now(pytz.utc))

                self._jobs.update_one(
                    {"_id": job["_id"]},
                    {"$set": {"next_run_time": next_run, "last_run_time": datetime.now(pytz.utc)}}
                )
            except Exception as e:
                self.logger.error(f"Task execution failed: {job['_id']} - {str(e)}")
                self._stop_lock_refresh(job["_id"])
                self._locks.delete_one({"job_id": job["_id"]})

    def _start_lock_refresh(self, job_id: str):
        stop_event = Event()

        refresh_thread = Thread(
            target=self._refresh_lock,
            args=(job_id, stop_event),
            daemon=True
        )
        refresh_thread.start()

        with self.lock_thread_lock:
            self.lock_refresh_threads[job_id] = stop_event

    def _stop_lock_refresh(self, job_id: str):
        with self.lock_thread_lock:
            stop_event = self.lock_refresh_threads.pop(job_id, None)
            if stop_event:
                stop_event.set()

    def _refresh_lock(self, job_id: str, stop_event: Event):
        while not stop_event.is_set():
            try:
                lock_expire = datetime.now(pytz.utc) + timedelta(seconds=self.lock_expire_seconds)
                self._locks.update_one(
                    {"job_id": job_id},
                    {"$set": {"expire_at": lock_expire}}
                )
                time.sleep(self.lock_refresh_interval)
            except Exception as e:
                self.logger.error(f"Failed to refresh lock for job {job_id}: {str(e)}")
                break

        try:
            self._locks.delete_one({"job_id": job_id})
        except Exception as e:
            self.logger.error(f"Failed to delete lock for job {job_id}: {str(e)}")

    @staticmethod
    def _run_task(func, args, kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            raise e

    def add_job(self, job_id, cron_str, func_object, timezone="UTC", args=None, kwargs=None, ipaddrs=None,
                exec_type="thread"):
        """
        Add a job to the scheduler.
        :param job_id: The ID of the job.
        :param cron_str: The cron string defining the schedule.
        :param func_object: The function object to execute.
        :param timezone: The timezone for the schedule.
        :param args: Additional arguments to pass to the function.
        :param kwargs: Additional keyword arguments to pass to the function.
        :param ipaddrs: List of allowed IP addresses to run the task.
        :param exec_type: The type of task to run. thread / process
        """

        existing_job = self._jobs.find_one({"_id": job_id})
        if existing_job:
            self.logger.warning(f"Job '{job_id}' already exists. Skipping insertion.")
            return

        cron_parser = CronParser(cron_str, timezone=timezone)
        next_run = cron_parser.get_next_run(datetime.now(pytz.utc))

        job_data = {
            "_id": job_id,
            "cron": cron_str,
            "func": pickle.dumps(func_object),
            "timezone": timezone,
            "args": args or [],
            "kwargs": kwargs or {},
            "enabled": True,
            "next_run_time": next_run,
            "last_run_time": None,
            "exec_type": exec_type
        }

        if ipaddrs is not None:
            job_data["ipaddrs"] = ipaddrs

        self._jobs.insert_one(job_data)

    def remove_job(self, job_id):
        """Remove a job from the scheduler."""
        self._jobs.delete_one({"_id": job_id})


def initialization_scheduler(config: dict):
    """
     Initialize the scheduler.
     :param config: The configuration dictionary.
    """
    Scheduler(config=config)


def scheduler_add_job(job_id, cron_str, func_object, timezone="UTC", args=None, kwargs=None, ipaddrs=None,
                      exec_type="thread"):
    """
    Add a job to the scheduler.
    :param job_id: The ID of the job.
    :param cron_str: The cron string defining the schedule.
    :param func_object: The function object to execute.
    :param timezone: The timezone for the schedule.
    :param args: Additional arguments to pass to the function.
    :param kwargs: Additional keyword arguments to pass to the function.
    :param ipaddrs: List of allowed IP addresses to run the task.
    :param exec_type: The type of task to run. thread / process
    """

    Scheduler().add_job(job_id, cron_str, func_object, timezone, args, kwargs, ipaddrs, exec_type)


def scheduler_remove_job(job_id):
    """
    Remove a job from the scheduler.
    :param job_id: The ID of the job to remove.
    """
    Scheduler().remove_job(job_id)


def scheduler_start():
    """
    Start the scheduler.
    """
    Scheduler().start()


def scheduler_stop():
    """
    Stop the scheduler.
    """
    Scheduler().stop()


def register():
    from astraflux.interface import scheduler
    scheduler.initialization_scheduler = initialization_scheduler
    scheduler.scheduler_add_job = scheduler_add_job
    scheduler.scheduler_remove_job = scheduler_remove_job
    scheduler.scheduler_start = scheduler_start
    scheduler.scheduler_stop = scheduler_stop

    import sys
    sys.modules['astraflux.interface.scheduler'] = scheduler



# -*- encoding: utf-8 -*-
import logging
from astraflux.inject import inject_implementation

__all__ = ['initialization_logger', 'loguru']


@inject_implementation()
def initialization_logger(config: dict):
    """
    Initialize the logger with the given configuration.
    Args:
        config (dict): A dictionary containing the configuration.
    """


@inject_implementation()
def loguru(filename: str, task_id: str = None) -> logging.Logger:
    """
    Get a logger instance for logging messages.
    Args:
        filename (str): The name of the log file.
        task_id (str, optional): The ID of the task. Defaults to None.
    Returns:
        logging.Logger: A logger instance.
    """



# -*- encoding: utf-8 -*-

import pika
import pickle
from functools import wraps

from astraflux.settings.keys import *
from astraflux.rpc.client import RpcClient

_RPC_METHODS = {}
_AMQP_CONFIG = {
    RABBITMQ.KEY_RABBITMQ_URI: RABBITMQ.DEFAULT_VALUE_RABBITMQ_URI,
    RPC.KEY_RPC_CALL_TIMEOUT: RPC.DEFAULT_VALUE_RPC_CALL_TIMEOUT
}


def _parse_config():
    """
    Parse the RabbitMQ configuration string to extract host, port, user and password.

    Returns:
        tuple: A tuple containing host (str), port (int), user (str) and password (str).
    """

    parts = _AMQP_CONFIG.get(RABBITMQ.KEY_RABBITMQ_URI).split('@')

    user_passwd = parts[0].split('//')[1]
    host_port = parts[1]

    user, passwd = user_passwd.split(':')
    host, port = host_port.split(':')

    return host, int(port), user, passwd


def _start_consumer(queue_name, service_instance):
    """
    Start a RabbitMQ consumer.
    Args:
        queue_name (str): The name of the queue to consume from.
        service_instance (function): The function to be called when a message is received.
    """
    host, port, user, passwd = _parse_config()
    credentials = pika.PlainCredentials(user, passwd)
    connection = pika.BlockingConnection(
        pika.ConnectionParameters(host=host, port=port, credentials=credentials, heartbeat=600)
    )
    channel = connection.channel()

    channel.queue_declare(
        queue=queue_name,
        durable=True,
        arguments={'x-ha-policy': 'all'}
    )
    channel.basic_qos(prefetch_count=100)

    def callback(ch, method_frame, props, body):
        response = None
        try:
            data = pickle.loads(body)
            method_name = data['method']
            args = data.get('args', [])
            kwargs = data.get('kwargs', {})

            if method_name not in _RPC_METHODS:
                raise AttributeError(f"Method {method_name} not registered as RPC")

            method = getattr(service_instance, method_name)
            result = method(*args, **kwargs)
            response = pickle.dumps({
                'status': 'success',
                'result': result
            })
        except Exception as e:
            response = pickle.dumps({
                'status': 'error',
                'exception': str(e),
                'type': type(e).__name__
            })
        finally:
            ch.basic_ack(method_frame.delivery_tag)
            ch.basic_publish(
                exchange='',
                routing_key=props.reply_to,
                properties=pika.BasicProperties(
                    correlation_id=props.correlation_id,
                    delivery_mode=2
                ),
                body=response
            )

    channel.basic_consume(
        queue=queue_name,
        on_message_callback=callback,
        consumer_tag=f"{queue_name}_consumer"
    )

    channel.start_consuming()


def initialization_rpc_proxy(config: dict):
    """
    Initialize the logger with the given configuration.
    Args:
        config (dict): A dictionary containing the configuration.
    """
    global _AMQP_CONFIG
    _AMQP_CONFIG[RABBITMQ.KEY_RABBITMQ_URI] = config.get(
        RABBITMQ.KEY_RABBITMQ_CONFIG, RABBITMQ.DEFAULT_VALUE_RABBITMQ_URI)

    _AMQP_CONFIG[RPC.KEY_RPC_CALL_TIMEOUT] = config.get(
        RPC.DEFAULT_VALUE_RPC_CALL_TIMEOUT, RPC.DEFAULT_VALUE_RPC_CALL_TIMEOUT)


def generate_unique():
    """
    Generates a unique identifier.
    Returns:
        str: The generated identifier.
    """
    client = RpcClient(config=_AMQP_CONFIG)
    service_name = '{}_{}_{}'.format(KEY_PROJECT_NAME, RPC.KEY_SYSTEM_SERVICE_NAME, 'task_distribution')
    data = client.call(service_name=service_name, method_name='generate_id')
    return data


def remote_call(service_name: str, method_name: str, **params):
    """
    Makes a remote procedure call to the specified service and method with the given parameters.

    Args:
        service_name (str): The name of the service to call.
        method_name (str): The name of the method to call.
        **params: Arbitrary keyword arguments to pass to the method.

    Returns:
        Any: The result of the remote procedure call.
    """
    client = RpcClient(config=_AMQP_CONFIG)
    data = client.call(
        service_name=service_name,
        method_name=method_name,
        **params
    )
    return data


def proxy_call(service_name: str, method_name: str, **params):
    """
    Makes a remote procedure call to the specified service and method with the given parameters.

    Args:
        service_name (str): The name of the service to call.
        method_name (str): The name of the method to call.
        **params: Arbitrary keyword arguments to pass to the method.

    Returns:
        Any: The result of the remote procedure call.
    """
    client = RpcClient(config=_AMQP_CONFIG)
    _name = '{}_{}'.format(KEY_PROJECT_NAME, service_name)

    data = client.call(
        service_name=_name,
        method_name=method_name,
        **params
    )
    return data


def rpc_decorator(func):
    """
    Decorator for RPC functions.
    Args:
        func (function): The function to be decorated.
    Returns:
        function: The decorated function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    global _RPC_METHODS
    _RPC_METHODS.setdefault(func.__name__)

    wrapper._is_rpc_func = True
    return wrapper


def service_running(service_cls, config):
    """
    Start a RabbitMQ consumer.
    Args:
        config (dict): The AMQP URL for the RabbitMQ server.
        service_cls (class): The function to be called when a message is received.
    """
    global _AMQP_CONFIG
    _AMQP_CONFIG = config

    service_instance = service_cls()
    service_name = getattr(service_instance, 'name', service_cls.__name__)
    _start_consumer(service_name, service_instance)


def register():
    from astraflux.interface import rpc
    rpc.initialization_rpc_proxy = initialization_rpc_proxy
    rpc.generate_unique = generate_unique
    rpc.remote_call = remote_call
    rpc.proxy_call = proxy_call
    rpc.rpc_decorator = rpc_decorator
    rpc.service_running = service_running

    import sys
    sys.modules['astraflux.interface.rpc'] = rpc



# -*- encoding: utf-8 -*-
import pymongo

from astraflux.settings.keys import *

_MONGODB_CONFIG = MONGODB.DEFAULT_VALUE_MONGODB_URI


class MongoClient:
    """
    This class is used to interact with a MongoDB database,
    providing a series of methods to operate on database collections.
    """
    _instance = None
    _table_name = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(MongoClient, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        self._client = pymongo.MongoClient(_MONGODB_CONFIG, connect=False)

    def _get_collection(self):
        """
        Get the MongoDB collection object for the specified table name.

        Returns:
            pymongo.collection.Collection: A MongoDB collection object.
        """
        return self._client[KEY_PROJECT_NAME][self._table_name]

    def update_many(self, query: dict, update_data: dict, upsert=False) -> None:
        """
        Update multiple documents in the collection that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            update_data (dict): A dictionary containing the data to update.
            upsert (bool, optional): If True, insert a new document if no documents match the query. Defaults to False.

        Returns:
            None
        """
        collection = self._get_collection()
        collection.update_many(query, {"$set": update_data}, upsert=upsert)
        return None

    def push_many(self, query: dict, update_data: dict, upsert=False) -> None:
        """
        Push data to an array field in multiple documents that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            update_data (dict): A dictionary containing the data to push.
            upsert (bool, optional): If True, insert a new document if no documents match the query. Defaults to False.

        Returns:
            None
        """
        collection = self._get_collection()
        collection.update_many(query, {"$push": update_data}, upsert=upsert)
        return None

    def push_one(self, query: dict, update_data: dict, upsert=False) -> None:
        """
        Push data to an array field in the first document that matches the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            update_data (dict): A dictionary containing the data to push.
            upsert (bool, optional): If True, insert a new document if no documents match the query. Defaults to False.

        Returns:
            None
        """
        collection = self._get_collection()
        collection.find_one_and_update(query, {"$push": update_data}, upsert=upsert)
        return None

    def pull_one(self, query: dict, update_data: dict) -> None:
        """
        Pull data from an array field in the first document that matches the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            update_data (dict): A dictionary containing the data to pull.

        Returns:
            None
        """
        collection = self._get_collection()
        collection.find_one_and_update(query, {"$pull": update_data})
        return None

    def insert_data(self, data: dict) -> None:
        """
        Insert a single document into the collection.

        Args:
            data (dict): A dictionary containing the data to insert.

        Returns:
            None
        """
        collection = self._get_collection()
        collection.insert_one(data)
        return None

    def delete_data(self, query: dict) -> None:
        """
        Delete multiple documents from the collection that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.

        Returns:
            None
        """
        collection = self._get_collection()
        collection.delete_many(query)
        return None

    def query_all(self, query: dict, field: dict) -> list[dict]:
        """
        Retrieve all documents from the collection that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            field (dict): A dictionary specifying the fields to include or exclude in the result.

        Returns:
            list: A list of documents that match the query.
        """
        collection = self._get_collection()
        data = collection.find(query, field)
        data = [i for i in data]
        return data

    def query_count(self, query) -> int:
        """
        Get the count of documents in the collection that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.

        Returns:
            int: The count of documents that match the query.
        """
        collection = self._get_collection()
        return collection.count_documents(filter=query)

    def query_list_sort(self, query: dict, field: dict, limit: int, skip_no: int,
                        sort_field: str = 'create_time', sort: int = -1) -> tuple[int, list[dict]]:
        """
        Retrieve a list of documents from the collection based on the given query,
        field projection, limit, skip, and sorting criteria.

        Args:
            query (dict): A dictionary specifying the query criteria.
            field (dict): A dictionary specifying the fields to include or exclude in the result.
            limit (int): The maximum number of documents to return.
            skip_no (int): The number of documents to skip before starting to return results.
            sort_field (str, optional): The field to sort the results by. Defaults to 'update_time'.
            sort (int, optional): The sorting order. -1 for descending, 1 for ascending. Defaults to -1.

        Returns:
            pymongo.cursor.Cursor: A cursor object that can be iterated over to access the documents.
        """
        collection = self._get_collection()
        data = collection.find(query, field).sort(sort_field, sort).limit(limit).skip(skip_no)
        data = [i for i in data]
        count = collection.count_documents(query)
        return count, data

    def query_one(self, query: dict, field: dict) -> list[dict]:
        """
        Retrieve a list of documents from the collection based on the given query,
        field projection, limit, skip, and sorting criteria.

        Args:
            query (dict): A dictionary specifying the query criteria.
            field (dict): A dictionary specifying the fields to include or exclude in the result.

        Returns:
            pymongo.cursor.Cursor: A cursor object that can be iterated over to access the documents.
        """
        collection = self._get_collection()
        data = collection.find_one(query, field)
        return data


def initialization_mongo(config: dict):
    """
    Initialize a MongoDB database connection with a MongoDB configuration.
    """
    global _MONGODB_CONFIG
    _MONGODB_CONFIG = config.get(MONGODB.KEY_MONGO_CONFIG, MONGODB.DEFAULT_VALUE_MONGODB_URI)


def mongodb_node() -> MongoClient:
    """
    Generate a MongoDB node with a MongoDB configuration.
    """

    class NODE(MongoClient):
        _table_name = TABLE.KEY_NODE_LIST

    return NODE()


def mongodb_task() -> MongoClient:
    """
    Generate a MongoDB task with a MongoDB configuration.
    """

    class TASK(MongoClient):
        _table_name = TABLE.KEY_TASK_LIST

    return TASK()


def mongodb_services() -> MongoClient:
    """
    Generate a MongoDB services with a MongoDB configuration.
    """

    class SERVICES(MongoClient):
        _table_name = TABLE.KEY_SERVICE_LIST

    return SERVICES()


def register():
    from astraflux.interface import mongodb
    mongodb.mongodb_node = mongodb_node
    mongodb.mongodb_task = mongodb_task
    mongodb.mongodb_services = mongodb_services
    mongodb.initialization_mongo = initialization_mongo

    import sys
    sys.modules['astraflux.interface.mongodb'] = mongodb



# -*- encoding: utf-8 -*-

import os
import logging
from logging.handlers import TimedRotatingFileHandler

from astraflux.settings import *

_FMT = logging.Formatter(LOG.DEFAULT_VALUE_LOGS_FMT)
_SUFFIX = LOG.DEFAULT_VALUE_LOGS_SUFFIX

_LOGS_POOL = {}
_LOGS_LEVEL = logging.INFO
_LOGS_FILE_PATH = os.path.expanduser("~")


def _get_log_file_path(filename: str, task_id: str = None):
    if task_id is None:
        return os.path.join(_LOGS_FILE_PATH, f'{filename}.log')
    else:
        os.makedirs(os.path.join(_LOGS_FILE_PATH, f'{filename}'), exist_ok=True)
        return os.path.join(_LOGS_FILE_PATH, filename, f'{task_id}.log')


def _get_logger(filename: str, task_id: str = None) -> logging.Logger:
    """
    Get a logger instance for logging messages.
    Args:
        filename (str): The name of the log file.
        task_id (str, optional): The ID of the task. Defaults to None.
    Returns:
        logging.Logger: A logger instance.
    """
    handler = _get_log_file_path(filename, task_id)
    if handler not in _LOGS_POOL:
        _logger = logging.getLogger(handler)
        _logger.setLevel(_LOGS_LEVEL)
        _logger.propagate = False

        th = TimedRotatingFileHandler(filename=handler, when='MIDNIGHT', backupCount=7, encoding='utf-8')
        th.suffix = _SUFFIX
        th.setFormatter(_FMT)

        if not any(isinstance(h, logging.StreamHandler) for h in _logger.handlers):
            ch = logging.StreamHandler()
            ch.setFormatter(_FMT)
            _logger.addHandler(ch)

        _logger.addHandler(th)

        _LOGS_POOL[handler] = _logger
    return _LOGS_POOL[handler]


def initialization_logger(config: dict):
    """
    Initialize the logger with the given configuration.
    Args:
        config (dict): A dictionary containing the configuration.
    """
    global _LOGS_FILE_PATH, _LOGS_LEVEL
    _LOGS_FILE_PATH = os.path.join(config.get(KEY_ROOT_PATH), LOG.KEY_LOGS_PATH)
    log_levels = {
        'DEBUG': logging.DEBUG,
        'INFO': logging.INFO,
        'WARNING': logging.WARNING,
        'ERROR': logging.ERROR
    }
    level = config.get(LOG.KEY_LOGS_CONFIG_NAME).get(LOG.KEY_LOGS_CONFIG_LEVEL)
    _LOGS_LEVEL = log_levels.get(level, logging.INFO)


def loguru(filename: str = None, task_id: str = None) -> logging.Logger:
    """
    Get a logger instance for logging messages.
    Args:
        filename (str): The name of the log file.
        task_id (str, optional): The ID of the task. Defaults to None.
    Returns:
        logging.Logger: A logger instance.
    """
    if task_id and filename:
        return _get_logger(filename, task_id)
    return _get_logger(filename=KEY_PROJECT_NAME, task_id=KEY_PROJECT_NAME)


def register():
    from astraflux.interface import logger
    logger.initialization_logger = initialization_logger
    logger.loguru = loguru

    import sys
    sys.modules['astraflux.interface.logger'] = logger



# -*- encoding: utf-8 -*-

import os
import sys
import inspect
import importlib
from typing import Union

from astraflux.settings import *
from astraflux.interface import *

__all__ = ['Build']


class Build:
    """
    Builds a worker by importing the specified class path and setting its attributes.
    """

    def __init__(self, config: dict, cls_path: str, build_type: str, constructor):
        self.config = config
        self.cls_path = cls_path
        self.build_type = build_type
        self.constructor = constructor

        initialization_mongo(config=config)
        initialization_redis(config=config)
        initialization_rabbitmq(config=config)
        initialization_rpc_proxy(config=config)
        initialization_logger(config=config)
        initialization_scheduler(config=config)

    def get_build_attr(self):
        """
        Imports the specified class path and returns its attributes.
        Returns:
            dict: A dictionary containing the attributes of the imported class.
        """
        if self.build_type == 'service':
            class_name = RPC.KEY_FUNCTION_RPC
        else:
            class_name = RPC.KEY_FUNCTION_WORKER

        script_path = os.path.dirname(self.cls_path)
        sys.path.insert(0, script_path)

        module_name, _file_extension = os.path.splitext(os.path.basename(self.cls_path))

        module = __import__(module_name, globals=globals(), locals=locals(), fromlist=[class_name])

        importlib.reload(module)
        cls = getattr(module, class_name)
        return cls.__dict__

    def build_functions(self, attrs):
        """
        Imports the specified class path and sets its attributes.
        """
        functions = {}
        for function_name in attrs:
            if function_name.startswith('__') is False:
                function = attrs[function_name]

                if type(function) in [type(lambda: None)]:
                    params = []
                    function = rpc_decorator(function)
                    signa = inspect.signature(function)
                    for name, param in signa.parameters.items():
                        if name != RPC.KEY_FUNCTION_SELF:
                            default_value = param.default
                            if param.default is inspect.Parameter.empty:
                                default_value = None

                            params.append({
                                RPC.KEY_FUNCTION_PARAM_NAME: name,
                                RPC.KEY_FUNCTION_PARAM_DEFAULT_VALUE: default_value
                            })

                    functions.setdefault(function_name, params)
                self.constructor.setattr(function_name, function)
        self.constructor.functions = functions

    def build(self, task_id: str = None) -> Union[ServiceConstructor, WorkerConstructor]:
        """
        Builds a worker by importing the specified class path and setting its attributes.
        Returns:
            ServiceConstructor: The constructed worker object.
        """
        attrs = self.get_build_attr()
        self.build_functions(attrs)

        self.constructor.subtask_create = subtask_create
        self.constructor.task_create = task_submit_databases
        self.constructor.task_create_and_run = task_submit_databases_and_send

        self.constructor.worker_ipaddr = get_ipaddr()
        self.constructor.service_ipaddr = get_ipaddr()
        self.constructor.worker_version = get_converted_time()
        self.constructor.service_version = get_converted_time()
        self.constructor.worker_name = attrs.get(BUILD.KEY_WORKER_NAME)
        self.constructor.service_name = attrs.get(BUILD.KEY_SERVICE_NAME)

        if self.build_type == 'service':
            self.constructor.name = '{}_{}'.format(KEY_PROJECT_NAME, self.constructor.service_name)

            self.constructor.loguru = loguru(filename=KEY_PROJECT_NAME, task_id=self.constructor.service_name)
        else:
            self.constructor.name = '{}_{}'.format(KEY_PROJECT_NAME, self.constructor.worker_name)

            if task_id is None:
                self.constructor.loguru = loguru(filename=KEY_PROJECT_NAME, task_id=self.constructor.worker_name)
            else:
                self.constructor.loguru = loguru(filename=self.constructor.worker_name, task_id=task_id)

        return self.constructor



# -*- encoding: utf-8 -*-
import pytz
import datetime

from astraflux.settings.keys import *

_fmt = TIME.DEFAULT_VALUE_TIME_FMT
_timezone = TIME.DEFAULT_VALUE_TIMEZONE


def _get_default_params(fmt, timezone):
    """
    Get default format and timezone if the input parameters are not provided.

    Args:
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        tuple: A tuple containing the format and timezone.
    """
    if fmt is False:
        fmt = _fmt

    if timezone is False:
        timezone = _timezone

    return fmt, timezone


def get_date_time_obj(data_str: str, fmt=False, timezone=False):
    """
    Specify the timezone and format, and return a time object.

    Args:
        data_str (str): The time string to be converted.
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        datetime.datetime: A datetime object representing the converted time.
    """
    fmt, timezone = _get_default_params(fmt, timezone)
    target_timezone = pytz.timezone(timezone)
    current_time = datetime.datetime.strptime(data_str, fmt)
    converted_time = current_time.astimezone(target_timezone)
    return converted_time


def format_converted_time(data_str: str, fmt=False, timezone=False, r_fmt=False):
    """
    Format a time string according to the specified format and timezone.

    Args:
        data_str (str): The time string to be formatted.
        fmt (str or bool): The input format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.
        r_fmt (str or bool): The output format of the time string. If False, use the input format.

    Returns:
        str: A formatted time string.
    """
    fmt, timezone = _get_default_params(fmt, timezone)
    if r_fmt is False:
        r_fmt = fmt

    current_time = datetime.datetime.strptime(data_str, fmt)
    target_timezone = pytz.timezone(timezone)
    converted_time = current_time.astimezone(target_timezone)

    return converted_time.strftime(r_fmt)


def get_converted_time(fmt=False, timezone=False):
    """
    Specify timezone and format, return the current time.

    Args:
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        str: A string representing the current time in the specified format and timezone.
    """
    fmt, timezone = _get_default_params(fmt, timezone)
    current_time = datetime.datetime.now()
    target_timezone = pytz.timezone(timezone)
    converted_time = current_time.astimezone(target_timezone)

    return converted_time.strftime(fmt)


def get_converted_time_float(fmt=False, timezone=False):
    """
    Specify timezone and format, return the current time as a float.
    Args:
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.
    Returns:
        float: A float representing the current time in the specified format and timezone.
    """
    return float(get_converted_time(fmt, timezone))


def get_yes_today(data_str: str, fmt=False, timezone=False):
    """
    Get the previous day's date from the given date string.

    Args:
        data_str (str): The date string.
        fmt (str or bool): The format of the date string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        str: A string representing the previous day's date in the specified format and timezone.
    """
    fmt, timezone = _get_default_params(fmt, timezone)

    day = datetime.datetime.strptime(data_str, fmt)
    delta = datetime.timedelta(days=-1)

    target_timezone = pytz.timezone(timezone)
    converted_time = day.astimezone(target_timezone)

    return (converted_time + delta).strftime(fmt)


def get_yesterday_date(fmt=False, timezone=False, days=1):
    """
    Specify timezone and format, return the previous day's time.

    Args:
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.
        days (int): The number of days to subtract from the current date. Default is 1.

    Returns:
        str: A string representing the previous day's time in the specified format and timezone.
    """
    fmt, timezone = _get_default_params(fmt, timezone)
    current_time = datetime.datetime.now()
    target_timezone = pytz.timezone(timezone)
    converted_time = current_time.astimezone(target_timezone)

    return (converted_time - datetime.timedelta(days=days)).strftime(fmt)


def convert_timestamp_to_timezone(timestamp, fmt=False, timezone=False):
    """
    Convert a timestamp to a time string in the specified timezone and format.

    Args:
        timestamp (float): The timestamp to be converted.
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        str: A string representing the converted time in the specified format and timezone.
    """
    fmt, timezone = _get_default_params(fmt, timezone)
    datetime_utc = datetime.datetime.utcfromtimestamp(timestamp)
    timezone_obj = pytz.timezone(timezone)
    datetime_timezone = datetime_utc.replace(tzinfo=pytz.utc).astimezone(timezone_obj)
    return datetime_timezone.strftime(fmt)


def get_converted_timestamp(date_string: str, fmt=False, timezone=False):
    """
    Convert a time string to a timestamp in the specified timezone and format.

    Args:
        date_string (str): The time string to be converted.
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.

    Returns:
        float: A timestamp representing the converted time.
    """
    fmt, timezone = _get_default_params(fmt, timezone)
    dt = datetime.datetime.strptime(date_string, fmt)
    target_tz = pytz.timezone(timezone)
    converted_dt = dt.astimezone(target_tz).replace(tzinfo=None)
    return converted_dt.timestamp()


def get_date_list(start_day: str, end_day: str):
    """
    Obtain every day within the time range.

    Args:
        start_day (str): The start date in the format '%Y%m%d'.
        end_day (str): The end date in the format '%Y%m%d'.

    Returns:
        list: A list of strings representing each day within the time range in the format '%Y%m%d'.
    """
    start_date = datetime.datetime.strptime(start_day, '%Y%m%d')
    end_date = datetime.datetime.strptime(end_day, '%Y%m%d')

    date_list = []
    current_date = start_date

    while current_date <= end_date:
        date_list.append(current_date.strftime('%Y%m%d'))
        current_date += datetime.timedelta(days=1)

    return date_list


def get_week_num(date_str: str):
    """
    Obtain a week number from the given date string.
    Args:
        date_str (str): The date string in the format '%Y-%m-%d'.
    Returns:
        int: The week number.
    """
    now_time = datetime.datetime.strptime(date_str + " 00:00:00", "%Y-%m-%d %H:%M:%S")
    one_time = now_time.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    week_num = int(now_time.strftime('%W')) - int(one_time.strftime('%W')) + 1
    return week_num


def get_current_week():
    """
    Obtain current week number.
    Returns:
        int: The week number.
    """
    today = datetime.date.today()
    monday = today - datetime.timedelta(days=today.weekday())
    return [datetime.datetime.strftime((monday + datetime.timedelta(days=i)), "%Y%m%d") for i in range(7)]


def is_timestamp_within_days(timestamp: int, fmt=False, timezone=False):
    """
    Check if a timestamp is within days.
    Args:
        timestamp (int): The timestamp to be checked.
        fmt (str or bool): The format of the time string. If False, use the default format.
        timezone (str or bool): The timezone. If False, use the default timezone.
    Returns:
        bool: True if the timestamp is within days, False otherwise.
    """
    fmt, timezone = _get_default_params(fmt, timezone)
    datetime_utc = datetime.datetime.utcfromtimestamp(timestamp)
    timezone_obj = pytz.timezone(timezone)
    datetime_timezone = datetime_utc.replace(tzinfo=pytz.utc).astimezone(timezone_obj)
    current_time = datetime.datetime.now()
    target_timezone = pytz.timezone(timezone)
    converted_time = current_time.astimezone(target_timezone)
    return (converted_time - datetime_timezone).days


def convert_timestamp_to_timezone_obj(timestamp, timezone=False):
    """
    Convert a timestamp to a time string.
    Args:
        timestamp (int): The timestamp to be converted.
        timezone (str or bool): The timezone. If False, use the default timezone.
    Returns:
        datetime.datetime: A datetime object representing the converted time.
    """
    fmt, timezone = _get_default_params(fmt=False, timezone=timezone)
    datetime_utc = datetime.datetime.utcfromtimestamp(timestamp)
    timezone_obj = pytz.timezone(timezone)
    datetime_timezone = datetime_utc.replace(tzinfo=pytz.utc).astimezone(timezone_obj)
    return datetime_timezone


def convert_timestamp_to_timezone_str(timestamp, timezone=False, fmt=False):
    """
    Convert a timestamp to a time string.
    Args:
        timestamp (int): The timestamp to be converted.
        timezone (str or bool): The timezone. If False, use the default timezone.
        fmt (str or bool): The format of the time string. If False, use the default format.
    Returns:
        str: A string representing the converted time in the specified format and timezone.
    """
    fmt, timezone = _get_default_params(fmt=fmt, timezone=timezone)
    datetime_utc = datetime.datetime.utcfromtimestamp(timestamp)
    timezone_obj = pytz.timezone(timezone)
    datetime_timezone = datetime_utc.replace(tzinfo=pytz.utc).astimezone(timezone_obj)
    return datetime_timezone.strftime(fmt)


def register():
    from astraflux.interface import format_time
    format_time.get_date_time_obj = get_date_time_obj
    format_time.format_converted_time = format_converted_time
    format_time.get_converted_time = get_converted_time
    format_time.get_yes_today = get_yes_today
    format_time.get_yesterday_date = get_yesterday_date
    format_time.convert_timestamp_to_timezone = convert_timestamp_to_timezone
    format_time.get_converted_timestamp = get_converted_timestamp
    format_time.get_date_list = get_date_list
    format_time.get_week_num = get_week_num
    format_time.get_current_week = get_current_week
    format_time.is_timestamp_within_days = is_timestamp_within_days
    format_time.convert_timestamp_to_timezone_obj = convert_timestamp_to_timezone_obj
    format_time.convert_timestamp_to_timezone_str = convert_timestamp_to_timezone_str
    format_time.get_converted_time_float = get_converted_time_float

    import sys
    sys.modules['astraflux.interface.format_time'] = format_time



# -*- encoding: utf-8 -*-
import os
import sys
import time
import json
import argparse
import multiprocessing

from astraflux.inject import inject_init
from astraflux.servers.build import Build

from astraflux.settings import *
from astraflux.interface import *


class TaskRun:

    @staticmethod
    def run(cls_path: str, config: dict, body: dict):

        initialization_mongo(config=config)
        initialization_redis(config=config)
        initialization_rabbitmq(config=config)
        initialization_rpc_proxy(config=config)
        initialization_logger(config=config)

        worker_pid = os.getpid()

        build = Build(
            config=config,
            cls_path=cls_path,
            build_type='worker',
            constructor=WorkerConstructor
        )
        constructor: WorkerConstructor = build.build(task_id=body[TASK.KEY_TASK_ID])

        start_ime = get_converted_time()
        mongodb_task().update_many(
            query={TASK.KEY_TASK_ID: body[TASK.KEY_TASK_ID]},
            update_data={
                BUILD.KEY_WORKER_PID: worker_pid,
                BUILD.KEY_WORKER_IPADDR: constructor.worker_ipaddr,
                TASK.KEY_TASK_STATUS: TASK.KEY_TASK_RUN_STATUS,
                TASK.KEY_TASK_START_TIME: start_ime
            }
        )

        mongodb_services().push_one(
            query={
                BUILD.KEY_WORKER_NAME: constructor.worker_name,
                BUILD.KEY_WORKER_IPADDR: constructor.worker_ipaddr
            },
            update_data={
                BUILD.KEY_WORKER_RUN_PROCESS: worker_pid
            }
        )

        try:
            constructor().run(body)
            mongodb_task().update_many(
                query={TASK.KEY_TASK_ID: body[TASK.KEY_TASK_ID]},
                update_data={
                    BUILD.KEY_WORKER_PID: os.getpid(),
                    BUILD.KEY_WORKER_IPADDR: constructor.worker_ipaddr,
                    TASK.KEY_TASK_STATUS: TASK.KEY_TASK_SUCCESS_STATUS,
                    TASK.KEY_TASK_END_TIME: get_converted_time()
                },
                upsert=True
            )
        except Exception as e:
            mongodb_task().update_many(
                query={TASK.KEY_TASK_ID: body[TASK.KEY_TASK_ID]},
                update_data={
                    BUILD.KEY_WORKER_PID: os.getpid(),
                    BUILD.KEY_WORKER_IPADDR: constructor.worker_ipaddr,
                    TASK.KEY_TASK_STATUS: TASK.KEY_TASK_ERROR_STATUS,
                    TASK.KEY_TASK_END_TIME: get_converted_time(),
                    TASK.KEY_TASK_ERROR_MESSAGE: str(e)
                },
                upsert=True
            )

        mongodb_services().pull_one(
            query={
                BUILD.KEY_WORKER_NAME: constructor.worker_name,
                BUILD.KEY_WORKER_IPADDR: constructor.worker_ipaddr
            },
            update_data={
                BUILD.KEY_WORKER_RUN_PROCESS: worker_pid
            }
        )


class RabbitmqCallback:
    """
    RabbitmqCallback is a class that represents a rabbitmq callback.
    It contains attributes such as name, config, loguru, ip_addr,
     cls_path, rpc_proxy, database_tasks, and database_services.
    """

    config = None
    loguru = None
    ip_addr = None
    cls_path = None
    worker_name = None

    def mq_callback(self, ch, method, properties, body):
        """
        Handles the callback for the rabbitmq message.
        Args:
            ch: The channel object.
            method: The method object.
            properties: The properties object.
            body: The body of the message.
        """
        ch.basic_ack(delivery_tag=method.delivery_tag)
        try:
            _body = json.loads(body.decode())
            if TASK.KEY_TASK_ID in _body:

                if BUILD.KEY_SYSTEM_SERVICE_NAME not in self.worker_name:
                    status = redis_get_task_status_by_task_id(task_id=_body.get(TASK.KEY_TASK_ID))
                else:
                    status = TASK.KEY_TASK_RUN_STATUS

                if status != TASK.KEY_TASK_STOP_STATUS:
                    run_worker, max_worker = query_worker_running_number(
                        query={
                            BUILD.KEY_WORKER_NAME: self.worker_name,
                            BUILD.KEY_SERVICE_IPADDR: self.ip_addr
                        }
                    )
                    if run_worker < max_worker:
                        multiprocessing.Process(target=TaskRun.run, args=(self.cls_path, self.config, _body,)).start()
                    else:
                        time.sleep(0.2)
                        ch.basic_publish(body=body, exchange='', routing_key=self.worker_name)
            else:
                self.loguru.error('{} is not find, error data : {}'.format(TASK.KEY_TASK_ID, _body))
        except Exception as e:
            self.loguru.error('mq_callback error: {}'.format(e))


class RunWorker:
    def __init__(self, config, cls_path):
        self.config = config
        self.cls_path = cls_path

    def worker_start(self):
        build = Build(
            config=self.config,
            cls_path=self.cls_path,
            build_type='worker',
            constructor=WorkerConstructor
        )
        constructor: WorkerConstructor = build.build(task_id=None)
        worker_data = {
            BUILD.KEY_NAME: constructor.name,
            BUILD.KEY_WORKER_IPADDR: constructor.worker_ipaddr,
            BUILD.KEY_WORKER_NAME: constructor.worker_name,
            BUILD.KEY_WORKER_VERSION: constructor.worker_version,
            BUILD.KEY_WORKER_PID: os.getpid(),
            BUILD.KEY_WORKER_FUNCTIONS: constructor.functions,
            BUILD.KEY_WORKER_MAX_PROCESS: 10,
            BUILD.KEY_WORKER_RUN_PROCESS: [],
        }

        mongodb_services().update_many(
            query={
                BUILD.KEY_SERVICE_IPADDR: constructor.worker_ipaddr,
                BUILD.KEY_SERVICE_NAME: constructor.worker_name
            },
            update_data=worker_data,
            upsert=True
        )
        constructor.loguru.info('Worker started == {}'.format(worker_data))

        mq_callback = RabbitmqCallback()
        mq_callback.config = self.config
        mq_callback.cls_path = self.cls_path
        mq_callback.loguru = constructor.loguru
        mq_callback.ip_addr = constructor.worker_ipaddr
        mq_callback.worker_name = constructor.worker_name

        while True:
            try:
                rabbitmq_receive_message(queue=constructor.worker_name, callback=mq_callback.mq_callback)
            except Exception as e:
                constructor.loguru.error(' {} work error : {}'.format(constructor.worker_name, e))
            time.sleep(0.5)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="run worker script")

    parser.add_argument("--config", type=str, help="worker config")
    parser.add_argument("--path", type=str, help="worker path")
    args = parser.parse_args()

    inject_init()

    configs = load_config(args.config)

    sys.path.append(configs[KEY_ROOT_PATH])

    RunWorker(config=configs, cls_path=args.path).worker_start()



# -*- encoding: utf-8 -*-

import json
import psutil
import platform

from astraflux.settings import *
from astraflux.interface import *


class PlatformInfo:
    """
    A class that retrieves and stores information about the operating system platform.
    """

    def __init__(self):
        """
        Initialize the PlatformInfo object.
        Retrieves system, release, version, machine, and processor information.
        """
        self.name = platform.system()
        self.release = platform.release()
        self.version = platform.version()
        self.machine = platform.machine()
        self.processor = platform.processor()

    def __str__(self):
        """
        Return a JSON string representation of the PlatformInfo object.
        """
        return json.dumps(self.__dict__)

    def __repr__(self):
        """
        Return a string representation of the PlatformInfo object.
        """
        return self.__str__()


class MemoryInfo:
    """
    A class that retrieves and stores information about the system's memory usage.
    """

    def __init__(self):
        """
        Initialize the MemoryInfo object.
        Retrieves total, available, used, and percentage of memory.
        """
        memory_info = psutil.virtual_memory()
        self.memory_total = round(memory_info.total / (1024 ** 3), 2)
        self.memory_available = round(memory_info.available / (1024 ** 3), 2)
        self.memory_used = round(memory_info.used / (1024 ** 3), 2)
        self.memory_percent = memory_info.percent

    def __str__(self):
        """
        Return a JSON string representation of the MemoryInfo object.
        """
        return json.dumps(self.__dict__)

    def __repr__(self):
        """
        Return a string representation of the MemoryInfo object.
        """
        return self.__str__()


class DiskInfo:
    """
    A class that retrieves and stores information about the system's disk usage.
    """

    def __init__(self):
        """
        Initialize the DiskInfo object.
        Retrieves total, used, free, and percentage of disk space.
        """
        disk_info = psutil.disk_usage('/')
        self.disk_total = round(disk_info.total / (1024 ** 3), 2)
        self.disk_used = round(disk_info.used / (1024 ** 3), 2)
        self.disk_free = round(disk_info.free / (1024 ** 3), 2)
        self.disk_percent = disk_info.percent

    def __str__(self):
        """
        Return a JSON string representation of the DiskInfo object.
        """
        return json.dumps(self.__dict__)

    def __repr__(self):
        """
        Return a string representation of the DiskInfo object.
        """
        return self.__str__()


class CPUInfo:
    """
    A class that retrieves and stores information about the system's CPU usage.
    """

    def __init__(self):
        """
        Initialize the CPUInfo object.
        Retrieves CPU count, percentage, maximum, minimum, and current frequency.
        """
        self.cpu_count = psutil.cpu_count()
        self.cpu_percent = psutil.cpu_percent(interval=1)
        self.cpu_freq_max = psutil.cpu_freq().max
        self.cpu_freq_min = psutil.cpu_freq().min
        self.cpu_freq_current = psutil.cpu_freq().current

    def __str__(self):
        """
        Return a JSON string representation of the CPUInfo object.
        """
        return json.dumps(self.__dict__)

    def __repr__(self):
        """
        Return a string representation of the CPUInfo object.
        """
        return self.__str__()


class NodeInfo:
    """
    A class that aggregates system information into a single node information object.
    """

    def __init__(self):
        """
        Initialize the _NodeInfo object.
        Retrieves node name, IP address, platform, memory, disk, CPU information, and update time.
        """
        self.name = platform.node()
        self.ipaddr = get_ipaddr()
        self.platform = PlatformInfo()
        self.memory = MemoryInfo()
        self.disk = DiskInfo()
        self.cpu = CPUInfo()

        self.update_time = get_converted_time('%Y-%m-%d %H:%M:%S')

        self.node = {
            'name': self.name,
            'ipaddr': self.ipaddr,
            'platform': self.platform.__dict__,
            'memory': self.memory.__dict__,
            'disk': self.disk.__dict__,
            'cpu': self.cpu.__dict__,
            'update_time': self.update_time
        }


class SystemMonitoring:
    """
    A class responsible for performing heartbeat detection.
    It continuously monitors the node's information and updates the database accordingly.
    """

    def __init__(self, config: dict):
        """
        Initialize the HeartbeatDetection instance.
        Args:
            config (dict): Configuration dictionary containing necessary settings.
        """
        initialization_logger(config=config)
        initialization_mongo(config=config)
        self.loguru = loguru(filename=KEY_PROJECT_NAME, task_id='system_monitoring')

    def run(self):
        """
            Start the heartbeat detection process.
            This method runs in an infinite loop, periodically checking the node's information
            and updating the database. If an exception occurs, it logs the error and continues.
        """
        node_info = NodeInfo()
        self.loguru.info(f"{node_info.node}")
        mongodb_node().update_many(
            query={SCHEDULE.KEY_NODE_IPADDR: node_info.ipaddr},
            update_data=node_info.node,
            upsert=True
        )



# -*- encoding: utf-8 -*-

import sys
import json
import yaml
import base64

from .keys import *

current_dir_ = 'None'
global_config_ = {}


def _gen_rabbitmq_uri(_config):
    """
    Generate rabbitmq uri
    """
    rabbitmq_uri = 'amqp://{}:{}@{}:{}'.format(
        _config['RabbitMQ']['username'],
        _config['RabbitMQ']['password'],
        _config['RabbitMQ']['host'],
        _config['RabbitMQ']['port']
    )
    _config[RABBITMQ.KEY_RABBITMQ_URI] = rabbitmq_uri


def _gen_mongodb_uri(_config):
    """
    Generate mongodb uri
    """
    mongodb_uri = 'mongodb://{}:{}@{}:{}'.format(
        _config['Mongodb']['username'],
        _config['Mongodb']['password'],
        _config['Mongodb']['host'],
        _config['Mongodb']['port']
    )
    _config[MONGODB.KEY_MONGODB_URI] = mongodb_uri


def _gen_redis_uri(_config):
    """
    Generate Redis uri
    """
    redis_uri = 'redis://:{}@{}:{}'.format(
        _config['Redis']['password'],
        _config['Redis']['host'],
        _config['Redis']['port']
    )
    _config[REDIS.KEY_REDIS_URI] = redis_uri


def _config_default_value_processing(config: dict):
    """
    Process the default values in the configuration dictionary.
    Args:
        config (dict): The configuration dictionary.
    Returns:
        dict: The configuration dictionary with processed default values.
    """

    if SCHEDULE.KEY_DEFAULT_SCHEDULE_TIME not in config:
        config[SCHEDULE.KEY_DEFAULT_SCHEDULE_TIME] = SCHEDULE.DEFAULT_VALUE_SCHEDULE_TIME

    return config


def current_dir() -> str:
    """
    Return the current directory of the current working directory.
    """
    return current_dir_


def global_config() -> dict:
    """
    Return the global configuration settings.
    """
    return global_config_


def load_settings(yaml_file: str, workspace: str) -> None:
    """
    Load settings from yaml file.
    :param yaml_file: yaml file path
    :param workspace: workspace path
    :return:
    """
    with open(yaml_file, "r", encoding="utf-8") as f:
        _config = yaml.safe_load(f)

    _gen_rabbitmq_uri(_config)
    _gen_mongodb_uri(_config)
    _gen_redis_uri(_config)

    _config[KEY_ROOT_PATH] = workspace

    global global_config_, current_dir_
    current_dir_ = workspace
    global_config_ = _config


def load_config(encoded_config: str) -> dict:
    """
    Decode and load the encoded configuration string.

    Args:
        encoded_config (str): The encoded configuration string.

    Returns:
        dict: The decoded configuration dictionary.
    """
    try:
        decoded_bytes = base64.b64decode(encoded_config)
        decoded_string = decoded_bytes.decode('utf-8')
        return _config_default_value_processing(json.loads(decoded_string))
    except Exception as e:
        print(f"Error loading config: {e}")
        sys.exit(1)


def dump_config(config: dict) -> str:
    """
    Dump the encoded configuration string.
    """
    json_str = json.dumps(config, ensure_ascii=False)
    json_bytes = json_str.encode('utf-8')
    return base64.b64encode(json_bytes).decode('utf-8')


def register():
    from astraflux.settings import config_processing
    config_processing.current_dir = current_dir
    config_processing.global_config = global_config
    config_processing.load_settings = load_settings
    config_processing.load_config = load_config
    config_processing.dump_config = dump_config

    import sys
    sys.modules['astraflux.settings.config_processing'] = config_processing



# -*- encoding: utf-8 -*-
import socket

from astraflux.settings.keys import *


def get_ipaddr() -> str:
    """
    Retrieves the IP address of the current machine by establishing a UDP connection
    to the specified IP and port.

    Returns:
        str: The IP address of the current machine.
    """
    socket_tools = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    socket_tools.connect((SOCKET.SOCKET_BIND_IP, SOCKET.SOCKET_BIND_PORT))
    return socket_tools.getsockname()[0]


def is_port_open(port: int, ip_addr: str = None) -> bool:
    """
    Checks if a specified port on a given IP address is open by attempting to establish
    a TCP connection.

    Args:
        ip_addr (str): The IP address to check.
        port (int): The port number to check.

    Returns:
        bool: True if the port is closed, False if the port is open.
    """
    if ip_addr is None:
        ip_addr = get_ipaddr()

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip_addr, int(port)))
        s.shutdown(SOCKET.SOCKET_SHUTDOWN_SLEEP)
        return False
    except IOError:
        return True


def register():
    from astraflux.interface import network
    network.get_ipaddr = get_ipaddr
    network.is_port_open = is_port_open

    import sys
    sys.modules['astraflux.interface.network'] = network



# -*- encoding: utf-8 -*-

import importlib

__all__ = ['inject_implementation']


def inject_implementation():
    """
    Inject NexusFlow implementation
    """

    def wrapper(func):
        def wrapped(*args, **kwargs):
            _module = importlib.import_module(func.__module__)
            _class = getattr(_module, func.__name__)
            return _class(*args, **kwargs)

        return wrapped

    return wrapper



# -*- encoding: utf-8 -*-
from typing import Callable, Optional, Dict, Any, List

from astraflux.inject import inject_implementation

__all__ = [
    "async_task_add",
    "async_task_run",
    "async_task_get_status",
    "async_task_wait",
    "async_task_stop_all",
    "async_task_list",
    "async_task_run_all",
    "async_task_wait_all"
]


@inject_implementation()
def async_task_add(
        task_type: str,
        target: Callable,
        task_id: Optional[str],
        args: tuple = (),
        kwargs: dict = None) -> str:
    """
    add task
    :param task_type: 'thread' / 'process'
    :param target:
    :param args:
    :param kwargs:
    :param task_id:
    :return: task_id
    """


@inject_implementation()
def async_task_run(task_id: str) -> bool:
    """run task"""


@inject_implementation()
def async_task_get_status(task_id: str) -> Optional[Dict[str, Any]]:
    """get task status"""


@inject_implementation()
def async_task_wait(task_id: str, timeout: Optional[float] = None) -> bool:
    """wait task"""


@inject_implementation()
def async_task_stop_all():
    """stop all tasks"""


@inject_implementation()
def async_task_list() -> Dict[str, Dict[str, Any]]:
    """get all tasks"""


@inject_implementation()
def async_task_run_all() -> List[str]:
    """
    run all tasks
    :return:
    """


@inject_implementation()
def async_task_wait_all(timeout: Optional[float] = None) -> bool:
    """
    wait all tasks
    :param timeout:
    :return:
    """



# -*- encoding: utf-8 -*-
from astraflux.inject import inject_implementation

__all__ = [
    "initialization_scheduler",
    "scheduler_add_job",
    "scheduler_remove_job",
    "scheduler_start",
    "scheduler_stop"
]


@inject_implementation()
def initialization_scheduler(config: dict):
    """
    Initialize the scheduler.
    :param config: The configuration dictionary.
    """


@inject_implementation()
def scheduler_add_job(job_id, cron_str, func_object, timezone="UTC", args=None, kwargs=None, ipaddrs=None,
                      exec_type="thread"):
    """
    Add a job to the scheduler.
    :param job_id: The ID of the job.
    :param cron_str: The cron string defining the schedule.
    :param func_object: The function object to execute.
    :param timezone: The timezone for the schedule.
    :param args: Additional arguments to pass to the function.
    :param kwargs: Additional keyword arguments to pass to the function.
    :param ipaddrs: List of allowed IP addresses to run the task.
    :param exec_type: The type of task to run. thread / process
    """


@inject_implementation()
def scheduler_remove_job(job_id):
    """
    Remove a job from the scheduler.
    :param job_id: The ID of the job to remove.
    """


@inject_implementation()
def scheduler_start():
    """
    Start the scheduler.
    """


@inject_implementation()
def scheduler_stop():
    """
    Stop the scheduler.
    """



# -*- encoding: utf-8 -*-

import pytz
import datetime
from calendar import monthrange

__all__ = ['CronParser']


class CronParser:

    def __init__(self, cron_str: str, timezone: str = pytz.utc):
        """
        Initializes the CronParser with a cron expression and an optional timezone.
        Args:
            cron_str (str): The cron expression string.
            timezone (Union[str, datetime.tzinfo], optional): The timezone to use for the cron expression.
                Defaults to pytz.utc.
        """

        self.cron_parts = cron_str.split()
        if len(self.cron_parts) != 6:
            raise ValueError("Invalid cron expression (Need 6 parts: second minute hour day month weekday)")

        self.timezone = timezone
        if isinstance(self.timezone, str):
            self.timezone = pytz.timezone(self.timezone)

        self.second = self._parse_part(self.cron_parts[0], 0, 59)
        self.minute = self._parse_part(self.cron_parts[1], 0, 59)
        self.hour = self._parse_part(self.cron_parts[2], 0, 23)
        self.day = self._parse_part(self.cron_parts[3], 1, 31)
        self.month = self._parse_part(self.cron_parts[4], 1, 12)
        self.weekday = self._parse_part(self.cron_parts[5], 0, 6)

    @staticmethod
    def _parse_part(part: str, min_val: int, max_val: int) -> set:
        """
        Parses a part of the cron expression and returns a set of valid values.
        Args:
            part (str): The part of the cron expression to parse.
            min_val (int): The minimum valid value for the part.
            max_val (int): The maximum valid value for the part.
        Returns:
            set: A set of valid values for the part.
        """
        if part == "*":
            return set(range(min_val, max_val + 1))
        values = set()
        for item in part.split(","):
            if "/" in item:
                step = int(item.split("/")[1])
                values.update(range(min_val, max_val + 1, step))
            elif "-" in item:
                start, end = map(int, item.split("-"))
                values.update(range(start, end + 1))
            else:
                values.add(int(item))
        return values

    def get_next_run(self, current_time: datetime.datetime):
        """
        Calculates the next run time based on the cron expression and the current time.
        Args:
            current_time (datetime): The current time.
        Returns:
            datetime: The next run time.
        """

        localized_time = current_time.astimezone(self.timezone)
        next_time = localized_time + datetime.timedelta(seconds=1)

        while True:

            if next_time.month not in self.month:
                next_time = next_time.replace(
                    day=1,
                    month=(next_time.month + 1) if next_time.month < 12 else 1,
                    year=(next_time.year + 1) if next_time.month == 12 else next_time.year,
                    hour=0, minute=0, second=0
                )
                continue

            day_valid = next_time.day in self.day
            weekday_valid = next_time.weekday() in self.weekday
            if not (day_valid or weekday_valid):
                next_time += datetime.timedelta(days=1)
                next_time = next_time.replace(hour=0, minute=0, second=0)
                continue

            if (
                    next_time.second in self.second
                    and next_time.minute in self.minute
                    and next_time.hour in self.hour
            ):
                return next_time.astimezone(pytz.utc)

            next_time += datetime.timedelta(seconds=1)

            if next_time.day > monthrange(next_time.year, next_time.month)[1]:
                next_time = next_time.replace(
                    day=1,
                    month=(next_time.month + 1) if next_time.month < 12 else 1,
                    year=(next_time.year + 1) if next_time.month == 12 else next_time.year,
                    hour=0, minute=0, second=0
                )



# -*- encoding: utf-8 -*-


KEY_PROJECT_VERSION = '1.0'

KEY_PROJECT_NAME = 'astraflux'
KEY_ROOT_PATH = 'current_dir'


class TABLE:
    KEY_TASK_LIST = 'task_list'
    KEY_NODE_LIST = 'node_list'
    KEY_SERVICE_LIST = 'service_list'


class LOG:
    KEY_LOGS_PATH = 'logs'
    KEY_LOGS_FILENAME = 'filename'
    KEY_LOGS_CONFIG_NAME = 'logger'
    KEY_LOGS_CONFIG_LEVEL = 'level'

    DEFAULT_VALUE_LOGS_SUFFIX = "%Y-%m-%d.log"
    DEFAULT_VALUE_LOGS_FMT = '%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] %(message)s'


class RABBITMQ:
    KEY_RABBITMQ_URI = 'AMQP_URI'
    KEY_RABBITMQ_CONFIG = 'RABBITMQ_CONFIG'
    KEY_SYSTEM_SERVICE_NAME = 'PSS'
    DEFAULT_VALUE_RABBITMQ_URI = 'amqp://scheduleAdmin:scheduleAdminPassword@127.0.0.1:5672'


class MONGODB:
    KEY_MONGODB_URI = 'MONGODB_URI'
    KEY_MONGO_CONFIG = 'MONGODB_CONFIG'
    DEFAULT_VALUE_MONGODB_URI = 'mongodb://scheduleAdmin:scheduleAdminPassword@127.0.0.1:27017'


class REDIS:
    KEY_REDIS_URI = 'REDIS_URI'
    KEY_REDIS_CONFIG = 'REDIS_CONFIG'
    DEFAULT_VALUE_REDIS_URI = 'redis://:scheduleAdminPassword@127.0.0.1:6379'


class RPC:
    KEY_RPC_CALL_TIMEOUT = 'RPC_CALL_TIMEOUT'
    DEFAULT_VALUE_RPC_CALL_TIMEOUT = 30
    KEY_SYSTEM_SERVICE_NAME = 'proxy'
    KEY_FUNCTION_SELF = 'self'
    KEY_FUNCTION_RPC = 'RpcFunction'
    KEY_FUNCTION_WORKER = 'WorkerFunction'
    KEY_FUNCTION_PARAM_NAME = 'param_name'
    KEY_FUNCTION_PARAM_DEFAULT_VALUE = 'default_value'


class SOCKET:
    SOCKET_BIND_PORT = 80
    SOCKET_BIND_IP = '8.8.8.8'
    SOCKET_SHUTDOWN_SLEEP = 2


class TIME:
    DEFAULT_VALUE_TIMEZONE = 'Asia/Shanghai'
    DEFAULT_VALUE_TIME_FMT = '%Y%m%d%H%M%S'


class BUILD:
    KEY_WORKER_PID = 'worker_pid'
    KEY_WORKER_NAME = 'worker_name'
    KEY_WORKER_IPADDR = 'worker_ipaddr'
    KEY_WORKER_VERSION = 'worker_version'
    KEY_WORKER_FUNCTIONS = 'worker_functions'
    KEY_WORKER_MAX_PROCESS = 'worker_max_process'
    KEY_WORKER_RUN_PROCESS = 'worker_run_process'

    KEY_NAME = 'name'
    KEY_SERVICE_PID = 'service_pid'
    KEY_SERVICE_NAME = 'service_name'
    KEY_SERVICE_IPADDR = 'service_ipaddr'
    KEY_SERVICE_VERSION = 'service_version'
    KEY_SERVICE_FUNCTIONS = 'service_functions'

    KEY_SYSTEM_SERVICE_NAME = 'proxy_system_server'


class TASK:
    DEFAULT_VALUE_TASK_WEIGHT = 1

    KEY_TASK_BODY = 'body'
    KEY_TASK_ID = 'task_id'
    KEY_TASK_WEIGHT = 'weight'
    KEY_TASK_STATUS = 'status'
    KEY_TASK_SOURCE_ID = 'source_id'
    KEY_TASK_QUEUE_NAME = 'queue_name'
    KEY_TASK_IS_SUB_TASK = 'is_subtask'
    KEY_TASK_IS_SUB_TASK_ALL_FINISH = 'is_subtask_all_finish'

    KEY_TASK_WAIT_STATUS = 'wait'
    KEY_TASK_SEND_STATUS = 'send'
    KEY_TASK_STOP_STATUS = 'stop'
    KEY_TASK_ERROR_STATUS = 'error'
    KEY_TASK_RUN_STATUS = 'running'
    KEY_TASK_SUCCESS_STATUS = 'success'
    KEY_TASK_ERROR_MESSAGE = 'error_message'

    KEY_TASK_END_TIME = 'end_time'
    KEY_TASK_START_TIME = 'start_time'
    KEY_TASK_CREATE_TIME = 'create_time'


class SCHEDULE:
    DEFAULT_VALUE_SCHEDULE_TIME = 10
    KEY_DEFAULT_SCHEDULE_TIME = 'DEFAULT_SCHEDULE_TIME'
    KEY_NODE_IPADDR = 'ipaddr'



# -*- encoding: utf-8 -*-

from astraflux.inject import inject_implementation

__all__ = [
    'initialization_nexusflow',
    'services_registry',
    'services_start'
]


@inject_implementation()
def initialization_nexusflow(config: dict):
    """
    initialization astraflux
    """


@inject_implementation()
def services_registry(services: list):
    """
    Registers a list of services to be managed.

    Args:
        services (list): A list of services to be registered.
    """


@inject_implementation()
def services_start():
    """
    Starts the service management process. Currently, this method is a placeholder
    and does not perform any actions.
    """



# -*- encoding: utf-8 -*-

from astraflux.settings import *
from astraflux.interface import *


class TaskDistribution:
    """
    This class is responsible for the task distribution logic within the system.
    It interacts with RabbitMQ and databases to manage tasks, filter them based on status,
    and distribute them to appropriate workers according to the worker's idle capacity.
    """

    def __init__(self, config: dict):
        self.config = config
        initialization_logger(config=config)
        initialization_mongo(config=config)
        self.loguru = loguru(filename=KEY_PROJECT_NAME, task_id='task_distribution')

    def run(self):

        query = {
            TASK.KEY_TASK_STATUS: {
                '$in': [TASK.KEY_TASK_WAIT_STATUS, TASK.KEY_TASK_RUN_STATUS, TASK.KEY_TASK_SUCCESS_STATUS]
            }
        }
        _, source_tasks = mongodb_task().query_list_sort(
            query=query, field={'_id': 0}, limit=1000, skip_no=0, sort_field=TASK.KEY_TASK_WEIGHT)

        self.loguru.info(f'source task count === {len(source_tasks)}')

        all_tasks = {}
        source_task_status = {}
        for task in source_tasks:
            task_id = task[TASK.KEY_TASK_ID]
            task_status = task[TASK.KEY_TASK_STATUS]
            queue_name = task[TASK.KEY_TASK_QUEUE_NAME]
            is_subtask = task[TASK.KEY_TASK_IS_SUB_TASK]

            if queue_name not in all_tasks:
                all_tasks[queue_name] = {
                    TASK.KEY_TASK_RUN_STATUS: {}, TASK.KEY_TASK_WAIT_STATUS: {}, TASK.KEY_TASK_SUCCESS_STATUS: {}}

            if is_subtask is False:
                source_task_status.setdefault(task_id, {'task_status': task_status, 'source_queue_name': queue_name})
                all_tasks[queue_name][task_status][task_id] = {'body': task, 'sub_tasks': []}

        for task in source_tasks:
            task_status = task[TASK.KEY_TASK_STATUS]
            queue_name = task[TASK.KEY_TASK_QUEUE_NAME]
            is_subtask = task[TASK.KEY_TASK_IS_SUB_TASK]
            source_id = task.get(TASK.KEY_TASK_SOURCE_ID)

            if task_status != TASK.KEY_TASK_WAIT_STATUS:
                continue

            if is_subtask:
                source_status = source_task_status[source_id]['task_status']
                source_queue_name = source_task_status[source_id]['source_queue_name']
                if source_status not in [TASK.KEY_TASK_STOP_STATUS, TASK.KEY_TASK_ERROR_STATUS]:
                    all_tasks[source_queue_name][source_status][source_id]['sub_tasks'].append(
                        {'body': task, 'queue_name': queue_name})

        run_tasks = {}
        sub_task_all_finish = {}
        for queue_name in all_tasks:
            if queue_name not in run_tasks:
                run_tasks[queue_name] = []

            for task_status in [TASK.KEY_TASK_RUN_STATUS, TASK.KEY_TASK_WAIT_STATUS, TASK.KEY_TASK_SUCCESS_STATUS]:
                for source_id in all_tasks[queue_name][task_status]:
                    task = all_tasks[queue_name][task_status][source_id]
                    sub_tasks = all_tasks[queue_name][task_status][source_id]['sub_tasks']

                    if task_status == TASK.KEY_TASK_WAIT_STATUS:
                        run_tasks[queue_name].append(task['body'])
                    else:
                        if len(sub_tasks) > 0:
                            for sub in sub_tasks:
                                sub_task = sub['body']
                                sub_task_queue_name = sub['queue_name']
                                if sub_task_queue_name not in run_tasks:
                                    run_tasks[sub_task_queue_name] = []

                                run_tasks[sub_task_queue_name].append(sub_task)
                        else:
                            sub_task_all_finish.setdefault(source_id, True)

        services = mongodb_services().query_all(
            query={
                BUILD.KEY_WORKER_NAME: {'$in': [i for i in run_tasks]}},
            field={
                '_id': 0,
                BUILD.KEY_WORKER_NAME: 1,
                BUILD.KEY_WORKER_MAX_PROCESS: 1,
                BUILD.KEY_WORKER_RUN_PROCESS: 1
            },
        )

        worker_idle_number = {}
        for service in services:
            worker_name = service[BUILD.KEY_WORKER_NAME]
            worker_max_process = service[BUILD.KEY_WORKER_MAX_PROCESS]
            worker_run_process = len(service[BUILD.KEY_WORKER_RUN_PROCESS])

            if worker_name not in worker_idle_number:
                worker_idle_number[worker_name] = 0
            worker_idle_number[worker_name] += worker_max_process - worker_run_process

        for queue_name in run_tasks:
            self.loguru.info(f'services {queue_name} wait task count == {len(run_tasks[queue_name])}')

            for task in run_tasks[queue_name]:
                if worker_idle_number[queue_name] > 0:
                    rabbitmq_send_message(queue=queue_name, message=task[TASK.KEY_TASK_BODY])
                    self.loguru.info(f'run task === {task[TASK.KEY_TASK_BODY]}')
                    worker_idle_number[queue_name] -= 1

        if len(sub_task_all_finish) > 0:
            mongodb_task().update_many(
                query={TASK.KEY_TASK_ID: {'$in': [i for i in sub_task_all_finish]}},
                update_data={TASK.KEY_TASK_IS_SUB_TASK_ALL_FINISH: True}
            )



# -*- encoding: utf-8 -*-
import json
import pika

from astraflux.settings.keys import *

__all__ = ['initialization_rabbitmq', 'rabbitmq_send_message', 'rabbitmq_receive_message']

_RABBITMQ_URI = RABBITMQ.DEFAULT_VALUE_RABBITMQ_URI
_MQ_CHANNEL = None


def _parse_config():
    """
    Parse the RabbitMQ configuration string to extract host, port, user and password.

    Returns:
        tuple: A tuple containing host (str), port (int), user (str) and password (str).
    """
    parts = _RABBITMQ_URI.split('@')
    user_passwd = parts[0].split('//')[1]
    host_port = parts[1]

    user, passwd = user_passwd.split(':')
    host, port = host_port.split(':')

    return host, int(port), user, passwd


def _mq_channel():
    """
    Create and return a new RabbitMQ channel.

    Returns:
        pika.channel.Channel: A new channel object for interacting with RabbitMQ.
    """
    global _MQ_CHANNEL
    if _MQ_CHANNEL:
        return _MQ_CHANNEL

    host, port, user, passwd = _parse_config()
    credentials = pika.PlainCredentials(user, passwd)
    connection = pika.BlockingConnection(pika.ConnectionParameters(
        host, port=port, virtual_host='/', credentials=credentials, heartbeat=0))
    _MQ_CHANNEL = connection.channel()
    return _MQ_CHANNEL


def _create_queue(queue: str):
    """
    Create a queue in RabbitMQ. If an error occurs during creation, log the error.

    Args:
        queue (str): The name of the queue to create.

    Returns:
        pika.channel.Channel: The channel object used to create the queue.
    """
    _channel = _mq_channel()
    while True:
        try:
            _channel.queue_declare(queue=queue)
            break
        except Exception as e:
            try:
                global _MQ_CHANNEL
                _MQ_CHANNEL = None

                _channel = _mq_channel()
                _MQ_CHANNEL = _channel
                _channel.queue_declare(queue=queue)
                break
            except Exception as e:
                print('create_queue name == ', queue, ' error == ', e)

    return _channel


def initialization_rabbitmq(config: dict):
    """
    Initialize the logger with the given configuration.
    Args:
        config (dict): A dictionary containing the configuration.
    """
    global _RABBITMQ_URI
    _RABBITMQ_URI = config.get(RABBITMQ.KEY_RABBITMQ_URI)


def rabbitmq_send_message(queue: str, message: dict):
    """
    Send a message to a specified queue in RabbitMQ.

    Args:
        queue (str): The name of the queue to send the message to.
        message (dict): The message to send. It will be converted to a JSON string if it's not already.
    """
    _channel = _create_queue(queue=queue)
    if not isinstance(message, str):
        message = json.dumps(message)
    _channel.basic_publish(exchange='', routing_key=queue, body=message)


def rabbitmq_receive_message(queue: str, callback):
    """
    Start consuming messages from a specified queue in RabbitMQ.

    Args:
        queue (str): The name of the queue to consume messages from.
        callback (callable): The callback function to handle received messages.
    """
    _channel = _create_queue(queue=queue)
    _channel.basic_consume(on_message_callback=callback, queue=queue, auto_ack=False)
    _channel.start_consuming()


def register():
    from astraflux.interface import rabbitmq
    rabbitmq.rabbitmq_send_message = rabbitmq_send_message
    rabbitmq.rabbitmq_receive_message = rabbitmq_receive_message
    rabbitmq.initialization_rabbitmq = initialization_rabbitmq

    import sys
    sys.modules['astraflux.interface.rabbitmq'] = rabbitmq



# -*- encoding: utf-8 -*-

from astraflux.inject import inject_implementation

__all__ = ['initialization_mongo', 'mongodb_node', 'mongodb_task', 'mongodb_services']


class MongoClient:
    """
    This class is used to interact with a MongoDB database,
    providing a series of methods to operate on database collections.
    """

    def update_many(self, query: dict, update_data: dict, upsert=False) -> None:
        """
        Update multiple documents in the collection that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            update_data (dict): A dictionary containing the data to update.
            upsert (bool, optional): If True, insert a new document if no documents match the query. Defaults to False.

        Returns:
            None
        """

    def push_many(self, query: dict, update_data: dict, upsert=False) -> None:
        """
        Push data to an array field in multiple documents that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            update_data (dict): A dictionary containing the data to push.
            upsert (bool, optional): If True, insert a new document if no documents match the query. Defaults to False.

        Returns:
            None
        """

    def push_one(self, query: dict, update_data: dict, upsert=False) -> None:
        """
        Push data to an array field in the first document that matches the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            update_data (dict): A dictionary containing the data to push.
            upsert (bool, optional): If True, insert a new document if no documents match the query. Defaults to False.

        Returns:
            None
        """

    def pull_one(self, query: dict, update_data: dict) -> None:
        """
        Pull data from an array field in the first document that matches the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            update_data (dict): A dictionary containing the data to pull.

        Returns:
            None
        """

    def insert_data(self, data: dict) -> None:
        """
        Insert a single document into the collection.

        Args:
            data (dict): A dictionary containing the data to insert.

        Returns:
            None
        """

    def delete_data(self, query: dict) -> None:
        """
        Delete multiple documents from the collection that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.

        Returns:
            None
        """

    def query_all(self, query: dict, field: dict) -> list[dict]:
        """
        Retrieve all documents from the collection that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.
            field (dict): A dictionary specifying the fields to include or exclude in the result.

        Returns:
            list: A list of documents that match the query.
        """

    def query_count(self, query) -> int:
        """
        Get the count of documents in the collection that match the query.

        Args:
            query (dict): A dictionary specifying the query criteria.

        Returns:
            int: The count of documents that match the query.
        """

    def query_list_sort(self, query: dict, field: dict, limit: int, skip_no: int,
                        sort_field: str = 'create_time', sort: int = -1) -> tuple[int, list[dict]]:
        """
        Retrieve a list of documents from the collection based on the given query,
        field projection, limit, skip, and sorting criteria.

        Args:
            query (dict): A dictionary specifying the query criteria.
            field (dict): A dictionary specifying the fields to include or exclude in the result.
            limit (int): The maximum number of documents to return.
            skip_no (int): The number of documents to skip before starting to return results.
            sort_field (str, optional): The field to sort the results by. Defaults to 'update_time'.
            sort (int, optional): The sorting order. -1 for descending, 1 for ascending. Defaults to -1.

        Returns:
            pymongo.cursor.Cursor: A cursor object that can be iterated over to access the documents.
        """

    def query_one(self, query: dict, field: dict) -> list[dict]:
        """
        Retrieve a list of documents from the collection based on the given query,
        field projection, limit, skip, and sorting criteria.

        Args:
            query (dict): A dictionary specifying the query criteria.
            field (dict): A dictionary specifying the fields to include or exclude in the result.

        Returns:
            pymongo.cursor.Cursor: A cursor object that can be iterated over to access the documents.
        """


@inject_implementation()
def initialization_mongo(config: dict):
    """
    Initialize a MongoDB database connection with a MongoDB configuration.
    """


@inject_implementation()
def mongodb_node() -> MongoClient:
    """
    Generate a MongoDB node with a MongoDB configuration.
    """


@inject_implementation()
def mongodb_task() -> MongoClient:
    """
    Generate a MongoDB task with a MongoDB configuration.
    """


@inject_implementation()
def mongodb_services() -> MongoClient:
    """
    Generate a MongoDB services with a MongoDB configuration.
    """



# -*- encoding: utf-8 -*-
from astraflux.inject import inject_implementation

__all__ = ['initialization_rabbitmq', 'rabbitmq_send_message', 'rabbitmq_receive_message']


@inject_implementation()
def initialization_rabbitmq(config: dict):
    """
    Initialize the logger with the given configuration.
    Args:
        config (dict): A dictionary containing the configuration.
    """


@inject_implementation()
def rabbitmq_send_message(queue: str, message: dict):
    """
    Send a message to a specified queue in RabbitMQ.

    Args:
        queue (str): The name of the queue to send the message to.
        message (dict): The message to send. It will be converted to a JSON string if it's not already.
    """


@inject_implementation()
def rabbitmq_receive_message(queue: str, callback):
    """
    Start consuming messages from a specified queue in RabbitMQ.

    Args:
        queue (str): The name of the queue to consume messages from.
        callback (callable): The callback function to handle received messages.
    """



# -*- encoding: utf-8 -*-

import sys
import time
import psutil
import subprocess
from pathlib import Path

from astraflux.settings import *
from astraflux.interface import *

_CONFIG = {}
_ENCODED_CONFIG = None
_SERVICE_LIST = []

_PYTHON_NAME = 'python'
_PYEXEC = sys.executable if 'python' in sys.executable else 'python3'


def _kill_process(script_file_path, target_file_path):
    """
    Kills the specified script file.
    Args:
        script_file_path : script file object
        target_file_path : target file object
    """

    for proc in psutil.process_iter(['name', 'cmdline']):
        try:
            name = proc.name()
            cmdline = proc.cmdline()
            if _PYTHON_NAME in name:
                script_file_path_proc = False
                target_file_path_proc = False
                for arg in cmdline:
                    if Path(script_file_path).as_posix() in Path(arg).as_posix():
                        script_file_path_proc = True

                    if Path(target_file_path).as_posix() in Path(arg).as_posix():
                        target_file_path_proc = True

                if script_file_path_proc and target_file_path_proc:
                    proc.kill()
                    time.sleep(0.1)
                    break
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            continue


def _start_server(service_main_file_path):
    """
    Starts the service and starts the heartbeat detection process.
    Args:
        service_main_file_path : The path to the service main file.
    Returns:
        int: The process ID of the newly started service process.
    """
    from . import run_server
    script_file = Path(run_server.__file__).resolve()
    _kill_process(script_file_path=script_file, target_file_path=service_main_file_path)

    command = [
        _PYEXEC,
        script_file,
        '--config', _ENCODED_CONFIG,
        '--path', service_main_file_path
    ]

    process = subprocess.Popen(command)
    return process.pid


def _start_worker(service_main_file_path):
    """
    Starts the service and starts the heartbeat detection process.
    Args:
        service_main_file_path : The path to the service main file.
    Returns:
        int: The process ID of the newly started service process.
    """
    from . import run_worker
    script_file = Path(run_worker.__file__).resolve()
    _kill_process(script_file_path=script_file, target_file_path=service_main_file_path)

    command = [
        _PYEXEC,
        script_file,
        '--config', _ENCODED_CONFIG,
        '--path', service_main_file_path
    ]
    process = subprocess.Popen(command)
    return process.pid


def _start_schedule():
    """
    Starts the schedule.
    Returns:
        int: The process ID of the newly started schedule process.
    """
    from astraflux.workflows import SystemMonitoring, TaskDistribution

    scheduler_add_job(
        job_id='system_monitoring',
        cron_str='0 0/1 * * * *',
        func_object=SystemMonitoring(config=_CONFIG).run
    )

    scheduler_add_job(
        job_id='task_distribution',
        cron_str='0/30 * * * * *',
        func_object=TaskDistribution(config=_CONFIG).run
    )
    scheduler_config = _CONFIG.get('scheduler')

    enable = True
    if scheduler_config:
        enable = scheduler_config.get('enable', False)

    if enable:
        scheduler_start()


def initialization_nexusflow(config: dict):
    """
    initialization astraflux
    """
    global _CONFIG, _ENCODED_CONFIG
    _CONFIG = config
    _ENCODED_CONFIG = dump_config(config=config)

    initialization_mongo(config=config)
    initialization_redis(config=config)
    initialization_rabbitmq(config=config)
    initialization_rpc_proxy(config=config)
    initialization_logger(config=config)
    initialization_scheduler(config=config)


def services_registry(services: list):
    """
    Registers a list of services to be managed.

    Args:
        services (list): A list of services to be registered.
    """

    [_SERVICE_LIST.append(i) for i in services]


def services_start():
    """
    Starts the service management process. Currently, this method is a placeholder
    and does not perform any actions.
    """
    _start_schedule()
    for service in _SERVICE_LIST:
        service_main_file_path = Path(service.__file__).resolve()
        _start_server(service_main_file_path=service_main_file_path)
        _start_worker(service_main_file_path=service_main_file_path)


def register():
    from astraflux.interface import run
    run.initialization_nexusflow = initialization_nexusflow
    run.services_registry = services_registry
    run.services_start = services_start

    import sys
    sys.modules['astraflux.interface.run'] = run



from setuptools import setup, find_packages

with open("README.rst", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='astraflux',
    version='1.0.1',
    description="AstraFlux Description",
    long_description=long_description,
    include_package_data=True,
    author='YanPing',
    author_email='zyphhxx@foxmail.com',
    maintainer='YanPing',
    maintainer_email='zyphhxx@foxmail.com',
    license='MIT License',
    url='https://gitee.com/ZYPH/astra-flux',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    python_requires=">=3.8",
    install_requires=[
        'psutil',
        'pika',
        'pytz',
        'pymongo',
        'pyyaml',
        'redis'
    ]
)



# -*- encoding: utf-8 -*-
from astraflux.inject import inject_implementation

__all__ = ['get_ipaddr', 'is_port_open']


@inject_implementation()
def get_ipaddr() -> str:
    """
    Retrieves the IP address of the current machine by establishing a UDP connection
    to the specified IP and port.

    Returns:
        str: The IP address of the current machine.
    """


@inject_implementation()
def is_port_open(port: int, ip_addr: str = None) -> bool:
    """
    Checks if a specified port on a given IP address is open by attempting to establish
    a TCP connection.

    Args:
        ip_addr (str): The IP address to check.
        port (int): The port number to check.

    Returns:
        bool: True if the port is closed, False if the port is open.
    """



# -*- encoding: utf-8 -*-

import os
import importlib

os_dir = os.path.dirname(os.path.dirname(__file__))


def get_last_dir(path):
    """
    Return the last directory of the given path.
    """
    normalized_path = os.path.normpath(path)
    head, tail = os.path.split(normalized_path)
    return tail if tail else os.path.basename(head)


def inject_init():
    for root, dirs, files in os.walk(os_dir):
        for file in files:
            if file.endswith('.py') and file.startswith('_') and file != '__init__.py':
                _module_name = 'astraflux.{}.{}'.format(get_last_dir(root), file[:-3])
                importlib.import_module(_module_name).register()
